
# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
"""
  Custom user-defined dynamically-evaluated quantities
"""
from .ValuedParam import ValuedParam, InputData, InputTypes

# class for custom dynamically-evaluated quantities
class Function(ValuedParam):
  """
    Represents a ValuedParam that takes values from a custom user-defined Python method.
  """

  @classmethod
  def get_input_specs(cls):
    """
      Template for parameters that can take a scalar, an ARMA history, or a function
      @ In, None
      @ Out, spec, InputData, value-based spec
    """
    spec = InputData.parameterInputFactory('Function', contentType=InputTypes.StringType,
        descr=r"""indicates this value should be taken from a Python function, as described
        in the \xmlNode{DataGenerators} node.""")
    spec.addParam('method', param_type=InputTypes.StringType,
        descr=r"""the name of the \xmlNode{DataGenerator} from which this value should be taken.""")
    return spec

  def __init__(self):
    """
      Constructor.
      @ In, None
      @ Out, None
    """
    super().__init__()
    self._method_name = None # name of the method within the module
    self._source_kind = 'Function'

  def read(self, comp_name, spec, mode, alias_dict=None):
    """
      Used to read valued param from XML input
      @ In, comp_name, str, name of component that this valued param will be attached to; only used for print messages
      @ In, spec, InputData params, input specifications
      @ In, mode, type of simulation calculation
      @ In, alias_dict, dict, optional, aliases to use for variable naming
      @ Out, needs, list, signals needed to evaluate this ValuedParam at runtime
    """
    super().read(comp_name, spec, mode, alias_dict=None)
    # aliases get used to convert variable names, notably for the cashflow's "capacity"
    if alias_dict is None:
      alias_dict = {}
    self._source_name = spec.value
    self._method_name = spec.parameterValues['method']
    return [self._method_name]

  def evaluate(self, inputs, target_var=None, aliases=None):
    """
      Evaluate this ValuedParam, wherever it gets its data from
      @ In, inputs, dict, stuff from RAVEN, particularly including the keys 'meta' and 'raven_vars'
      @ In, target_var, str, optional, requested outgoing variable name if not None
      @ In, aliases, dict, optional, alternate variable names for searching in variables
      @ Out, data, dict, dictionary of resulting evaluation as {vars: vals}
      @ Out, meta, dict, dictionary of meta (possibly changed during evaluation)
    """
    if aliases is None:
      aliases = {}
    # TODO how to handle aliases for functions?
    # the "request" is what we're asking for from the function, the first argument given.
    # -> note it can be None if the function is not a transfer-type function
    request = inputs.pop('request', None)
    data, meta = self._target_obj.evaluate(self._method_name, request, inputs)
    return data, meta




# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
"""
  utilities for use within heron
"""
import sys
import importlib
import xml.etree.ElementTree as ET
import warnings
import pickle
try:
  from functools import cache
except ImportError:
  from functools import lru_cache
  def cache(user_function):
     """
       use lru_cache for older versions of python
       @ In, user_function, function
       @ Out, user_function, function that caches values
     """
     return lru_cache(maxsize=None)(user_function)
from os import path

import pandas as pd


def get_heron_loc():
  """
    Return HERON location
    @ In, None
    @ Out, loc, string, absolute location of HERON
  """
  return path.abspath(path.join(__file__, '..', '..'))

def get_raven_loc():
  """
    Return RAVEN location
    hopefully this is read from heron/.ravenconfig.xml
    @ In, None
    @ Out, loc, string, absolute location of RAVEN
  """
  try:
    import ravenframework
    print("WARNING: get_raven_loc deprecated")
    import traceback
    traceback.print_stack()
    return path.dirname(ravenframework.__path__[0])
  except ModuleNotFoundError:
    pass
  config = path.abspath(path.join(path.dirname(__file__),'..','.ravenconfig.xml'))
  if not path.isfile(config):
    raise IOError(
        f'HERON config file not found at "{config}"! Has HERON been installed as a plugin in a RAVEN installation?'
    )
  loc = ET.parse(config).getroot().find('FrameworkLocation')
  assert loc is not None and loc.text is not None
  # The addition of ravenframework as an installable package requires
  # adding the raven directory to the PYTHONPATH instead of adding
  # ravenframework. We will expect '.ravenconfig.xml' to point to
  # raven/ravenframework always, so this is why we grab the parent dir.
  return path.abspath(path.dirname(loc.text))

def get_cashflow_loc(raven_path=None):
  """
    Get CashFlow (aka TEAL) location in installed RAVEN
    @ In, raven_path, string, optional, if given then start with this path
    @ Out, cf_loc, string, location of CashFlow
  """
  if raven_path is None:
    raven_path = get_raven_loc()
  plugin_handler_dir = path.join(raven_path, 'scripts')
  sys.path.append(plugin_handler_dir)
  sys.path.append(path.join(raven_path, 'scripts'))
  plugin_handler = importlib.import_module('plugin_handler')
  sys.path.pop()
  sys.path.pop()
  cf_loc = plugin_handler.getPluginLocation('TEAL')
  return cf_loc

def get_farm_loc(raven_path=None): # Added by Haoyu Wang, May 25, 2022
  """
    Get FARM location in installed RAVEN
    @ In, raven_path, string, optional, if given then start with this path
    @ Out, cf_loc, string, location of CashFlow
  """
  if raven_path is None:
    raven_path = get_raven_loc()
  plugin_handler_dir = path.join(raven_path, 'scripts')
  sys.path.append(plugin_handler_dir)
  plugin_handler = importlib.import_module('plugin_handler')
  sys.path.pop()
  farm_loc = plugin_handler.getPluginLocation('FARM')
  return farm_loc

def get_all_resources(components):
  """
    Provides a set of all resources used among all components
    @ In, components, list, HERON component objects
    @ Out, resources, list, resources used in case
  """
  res = set()
  for comp in components:
    res.update(comp.get_resources())
  return res

def get_project_lifetime(case, components):
  """
    obtains the project lifetime
    @ In, case, HERON case, case
    @ In, components, list, HERON components
    @ Out, lifetime, int, project lifetime (usually years)
  """
  # load CashFlow
  try:
    from TEAL.src.main import getProjectLength
    from TEAL.src import CashFlows
  except (ImportError, ModuleNotFoundError) as e:
    loc = get_cashflow_loc()
    sys.path.append(loc)
    from TEAL.src.main import getProjectLength
    from TEAL.src import CashFlows
    sys.path.pop()
  econ_comps = list(comp.get_economics() for comp in components)
  econ_params = case.get_econ(econ_comps)
  econ_settings = CashFlows.GlobalSettings()
  econ_settings.setParams(econ_params)
  return getProjectLength(econ_settings, econ_comps)

@cache
def get_synthhist_structure(fpath):
  """
    Extracts synthetic history info from ROM (currently ARMA ROM)
    @ In, fpath, str, full absolute path to serialized ROM
    @ Out, structure, dict, derived structure from reading ROM XML
  """
  # TODO could this be a function of the ROM itself?
  # TODO or could we interrogate the ROM directly instead of the XML?
  try:
    import ravenframework
  except ModuleNotFoundError:
    #If ravenframework not in path, need to add, otherwise loading rom will fail
    raven_path = hutils.get_raven_loc()
    sys.path.append(os.path.expanduser(raven_path))
  rom = pickle.load(open(fpath, 'rb'))

  structure = {}
  meta = rom.writeXML().getRoot()
  # interpolation information
  interp_node = meta.find('InterpolatedMultiyearROM')
  if interp_node:
    macro_id = interp_node.find('MacroParameterID').text.strip()
    structure['macro'] = {
      'id': macro_id,
      'num': int(interp_node.find('MacroSteps').text),
      'first': int(interp_node.find('MacroFirstStep').text),
      'last': int(interp_node.find('MacroLastStep').text),
    }
    macro_nodes = meta.findall('MacroStepROM')
  else:
    macro_nodes = [meta]

  # cluster information
  structure['clusters'] = {}
  for macro in macro_nodes:
    if interp_node:
      macro_index = int(macro.attrib[macro_id])
    else:
      macro_index = 0
    clusters_info = [] # data dict for each macro step
    structure['clusters'][macro_index] = clusters_info
    cluster_nodes = macro.findall('ClusterROM')
    if cluster_nodes:
      for node in cluster_nodes:
        info = {
          'id': int(node.attrib['cluster']),
          'represents': node.find('segments_represented').text.split(','),
          'indices': list(int(x) for x in node.find('indices').text.split(','))
        }
        clusters_info.append(info)
  # segment information
  # -> TODO
  structure['segments'] = {}
  return structure

def get_csv_structure(fpath, macro_var, micro_var):
  """
    Returns CSV structure in a way RAVEN & HERON understand
    @ In, fpath, str, file path to CSV file
    @ In, macro_var, str, Macro Variable name - typically 'Year'
    @ In, micro_var, str, Micro Variable name - typically 'Time'
    @ Out, structure, dict, Nested structure of the CSV dataframe.
  """
  data = pd.read_csv(fpath)
  structure = {}
  if macro_var in data.columns:
    macro_steps = pd.unique(data[macro_var].values)
    structure['macro'] = {
      'id': macro_var,
      'num': len(macro_steps) + 1,
      'first': min(macro_steps),
      'last': max(macro_steps)
    }
  else:
    data[macro_var] = 0

  structure['clusters'] = {}
  for macro_step, df in data.groupby(macro_var):
    structure['clusters'][macro_step] = [{
      'id': 0,
      'indices': [0, len(df[micro_var].values)],
      'represents': ['0']
    }]

  structure['segments'] = {}  # TODO
  return structure


if __name__ == '__main__':
  try:
    action = sys.argv[1]
  except IndexError:
    print('Please provide the method to run as the first argument.')
  if action == 'get_raven_loc':
    print(get_raven_loc())
  elif action == 'get_cashflow_loc':
    print(get_cashflow_loc())
  else:
    raise IOError('Unrecognized action: "{}"'.format(action))




# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
"""
  Base classes for other classes.
"""
import os
import sys

import HERON.src._utils as hutils
try:
  import ravenframework
except ModuleNotFoundError:
  raven_path = hutils.get_raven_loc()
  sys.path.append(os.path.expanduser(raven_path))
from ravenframework.BaseClasses import MessageUser


class Base(MessageUser):
  def __init__(self, **kwargs):
    """
      Constructor.
      @ In, kwargs, dict, passthrough arguments
      @ Out, None
    """
    super().__init__()
    self.messageHandler = kwargs.get('messageHandler', None)

  def __repr__(self):
    """
      String representation.
      @ In, None
      @ Out, repr, str rep
    """
    return '<HERON {}>'.format(self.__class__.__name__)

  def set_message_handler(self, mh):
    """
      Sets message handling tool.
      @ In, mh, MessageHandler, message handler
      @ Out, None
    """
    self.messageHandler = mh

  def print_me(self, **kwargs):
    """
      Prints info about self.
      @ In, kwargs, dict, passthrough args
      @ Out, None
    """
    print('For {} no printing has been implemented!'.format(self))



# Copyright 2022, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
"""
  Alternative analysis approach to HERON's standard RAVEN running RAVEN, contains all the necessary methods to run
  a monolithic solve that utilizes TEAL cashflows, RAVEN ROM(s), and pyomo optimization.
"""
import os
import sys
from functools import partial
import itertools as it

import pyomo.environ as pyo
from pyomo.opt import SolverFactory
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from HERON.src import _utils as hutils
from HERON.src.base import Base
try:
  import ravenframework
except ModuleNotFoundError:
  path_to_raven = hutils.get_raven_loc()
  sys.path.append(os.path.abspath(os.path.join(path_to_raven, 'plugins')))
  sys.path.append(path_to_raven)
from TEAL.src import main as RunCashFlow
from TEAL.src import CashFlows
from ravenframework.ROMExternal import ROMLoader
from ravenframework.MessageHandler import MessageHandler

class MOPED(Base):
  def __init__(self):
    """
      Construct.
      @ In, None
      @ Out, None
    """
    super().__init__()
    self._components = []                 # List of components objects from heron input
    self._sources = []                    # List of sources objects from heron input
    self._case = None                     # Case object that contains the case parameters
    self._econ_settings = None            # TEAL global settings used for building cashflows
    self._m = None                        # Pyomo model to be solved
    self._producers = []                  # List of pyomo var/params of producing components
    self._eval_mode = 'clustered'         # (full or clustered) clustered is better for testing and speed, full gives a more realistic NPV result
    self._yearly_hours = 24 * 365         # Number of hours in a year to handle dispatch, based on clustering
    self._component_meta = {}             # Primary data structure for MOPED, organizes important information for problem construction
    self._cf_meta = {}                    # Secondary data structure for MOPED, contains cashflow info
    self._resources = []                  # List of resources used in this analysis
    self._solver = SolverFactory('ipopt') # Solver for optimization solve, default is 'ipopt', TODO allow user to specify
    self._cf_components = []              # List of TEAL.Components objects generated for analysis
    self._dispatch = []                   # List of pyomo vars/params for each realization and year
    self._multiplicity_meta = {}          # Dictionary of analysis years, clusters, and associated multiplicity
    self._plot = False                    # Boolean to determine if a dispatch plot is made for the analysis (defaults to false)

    self.messageHandler = MessageHandler()

  def buildActivity(self):
    """
      Generate active list that is necessary for building TEAL settings object
      @ In, None
      @ Out, activity, list, associates component with cashflow types ([ngcc|'Cap', ngcc|'Hourly'])
    """
    activity = []
    for comp in self._components:
      # TODO Does this need to be expanded on?
      for cf in comp._economics._cash_flows:
        if cf._type == 'one-time':
          activity.append(f'{comp.name}|Capex')
        elif cf._type == 'repeating':
          if cf._period == 'year':
            activity.append(f'{comp.name}|Yearly')
            continue
          # Necessary to have activity indicator account for multiple dispatch realizations
          for real in range(self._case._num_samples):
            activity.append(f'{comp.name}|Dispatching_{real+1}')
    self.raiseADebug(f'Built activity Indicator: {activity}')
    return activity

  def buildEconSettings(self, verbosity=0):
    """
      Builds TEAL economic settings for running cashflows
      @ In, verbosity, int or string, verbosity settings for TEAL
      @ out, None
    """
    activity = self.buildActivity()
    params = self._case._global_econ
    params['Indicator']['active'] = activity
    params['verbosity'] = verbosity
    self.raiseADebug('Building economic settings...')
    valid_params = ['ProjectTime', 'DiscountRate',
                    'tax', 'inflation', 'verbosity', 'Indicator']
    for param_name, param_value in params.items():
      if param_name != 'Indicator' and param_name in valid_params:
        self.raiseADebug(f'{param_name}: {param_value}')
      elif param_name == 'Indicator':
        self.raiseADebug(f'Indicator dictionary: {params["Indicator"]}')
      else:
        raise IOError(f'{param_name} is not a valid economic setting')
    self.raiseADebug('Finished building economic settings!')
    self._econ_settings = CashFlows.GlobalSettings()
    self._econ_settings.setParams(params)

  def buildComponentMeta(self):
    """
      Build pyomo object, capacity variables, and fixed capacity parameters
      @ In, None
      @ Out, None
    """
    self._m = pyo.ConcreteModel(name=self._case.name)
    # Considering all components in analysis to build a full pyomo solve
    for comp in self._components:
      self._component_meta[comp.name] = {}
      for prod in comp._produces:  # NOTE Cannot handle components that produce multiple things
        resource = prod._capacity_var
        mode = prod._capacity.type
        self.setCapacityMeta(mode, resource, comp, prod, kind='produces')
      for dem in comp._demands:  # NOTE Cannot handle components that demand multiple things
        resource = dem._capacity_var
        mode = dem._capacity.type
        self.setCapacityMeta(mode, resource, comp, dem, kind='demands')
      for store in comp._stores:  # NOTE Cannot handle components that store multiple things
        resource = store._capacity_var
        mode = store._capacity.type
        self.setCapacityMeta(mode, resource, comp, store, kind='stores')

  def buildMultiplicityMeta(self):
    """
      Loads source structure and builds appropriate multiplicity data
      @ In, None
      @ Out, None
    """
    structure = hutils.get_synthhist_structure(self._sources[0]._target_file)
    cluster_years = sorted(structure['clusters'])
    for i in range(len(cluster_years)):
      self._multiplicity_meta[i+1] = {}
      # Necessary to still allow full eval mode
      if self._eval_mode == 'full':
        self._multiplicity_meta[i+1][0] = 1
        continue
      cluster_data = structure['clusters'][cluster_years[i]]
      for cluster_info in cluster_data:
        self._multiplicity_meta[i+1][cluster_info['id']] = len(cluster_info['represents'])
    self._multiplicity_meta['Index Map'] = '[Year][Cluster][Multiplicity]'

  def loadSyntheticHistory(self, signal, multiplier):
    """
      Loads synthetic history for a specified signal,
      also sets yearly hours and pyomo indexing sets
      @ In, signal, string, name of signal to sample
      @ In, multiplier, int/float, value to multiply synthetic history evaluations by
      @ Out, synthetic_data, dict, contains data from evaluated ROM
    """
    # NOTE self._sources[0]._var_names are the user assigned signal names in DataGenerators
    if signal not in self._sources[0]._var_names:
      raise IOError('The requested signal name is not available'
                    'from the synthetic history, check DataGenerators node in input')
    runner = ROMLoader(self._sources[0]._target_file)
    from ravenframework.utils import xmlUtils
    inp = {'scaling': [1]}
    # TODO expand to change other pickledROM settings withing this method
    nodes = []
    node = xmlUtils.newNode('ROM', attrib={'name': 'SyntheticHistory', 'subType': 'pickledRom'})
    node.append(xmlUtils.newNode('clusterEvalMode', text=self._eval_mode))
    nodes.append(node)
    runner.setAdditionalParams(nodes)
    synthetic_data = {}
    for real in range(self._case._num_samples):
      self.raiseAMessage(f'Loading synthetic history for signal: {signal}')
      name = f'Realization_{real + 1}'
      current_realization = runner.evaluate(inp)[0]
      # applying mult to each realization is easier than iteration through dict object later
      current_realization[signal] *= multiplier
      if self._eval_mode == 'full':
        # reshape so that a filler cluster index is made
        current_realization[signal] = np.expand_dims(current_realization[signal], axis=1)
      synthetic_data[name] = current_realization[signal]
    cluster_count = synthetic_data['Realization_1'].shape[1]
    hour_count = synthetic_data['Realization_1'].shape[2]
    self._m.c = pyo.Set(initialize=np.arange(cluster_count))
    if self._eval_mode not in ['clustered', 'full']:
      raise IOError('Improper ROM evaluation mode detected, try "clustered" or "full".')
    # How many dispatch points we will have for each year
    self._yearly_hours = hour_count * cluster_count
    self._m.t = pyo.Set(initialize=np.arange(hour_count))
    return synthetic_data

  def setCapacityMeta(self, mode, resource, comp, element, kind='produces'):
    """
      Checks the capacity type, dispatch type, and resources involved for each component
      to build component_meta
      @ In, mode, string, type of capacity definition for component
      @ In, resource, string, resource produced or demanded
      @ In, comp, HERON component
      @ In, element, HERON produces/demands node
      @ In, kind, string, describes comp type, stores,produces,demands
      @ Out, None
    """
    # Multiplier plays important role in capacity node, especially for VRE's
    capacity_mult = element._capacity._multiplier
    # For MOPED we treat all capacities and dispatches as positive values
    if capacity_mult is None:
      capacity_mult = 1
    elif capacity_mult < 0:
      capacity_mult *= -1
    self._component_meta[comp.name]['Capacity Resource'] = resource
    self._component_meta[comp.name]['Stores'] = None
    self._component_meta[comp.name]['Produces'] = None
    self._component_meta[comp.name]['Demands'] = None
    # Organizing important aspects of problem for later access
    # TODO considering lists of produce and demand
    if kind == 'produces':
      self._component_meta[comp.name]['Produces'] = element._produces[0]
    elif kind == 'demands':
      self._component_meta[comp.name]['Demands'] = element._demands[0]
    elif kind == 'stores':
      self._component_meta[comp.name]['Stores'] = element._stores
      self._component_meta[comp.name]['Initial Value'] = element._initial_stored.get_value()
      self._component_meta[comp.name]['SRTE'] = element.get_sqrt_RTE()
    self._component_meta[comp.name]['Consumes'] = None
    self._component_meta[comp.name]['Dispatch'] = element._dispatchable
    # Different possible capacity value definitions for a component
    if mode == 'OptBounds':
      self.raiseADebug(f'Building pyomo capacity variable for '
                       f'{comp.name}')
      opt_bounds = element._capacity._vp._parametric
      # Considering user inputs for default heron sign convention
      if opt_bounds[1] < 1:
        opt_bounds *= -1
      opt_bounds *= capacity_mult
      # This is a capacity we make a decision on
      var = pyo.Var(initialize=0.5 * opt_bounds[1], bounds=(opt_bounds[0], opt_bounds[1]))
      setattr(self._m, f'{comp.name}', var)
    elif mode == 'SweepValues':  # TODO Add capability to handle sweepvalues, maybe multiple pyo.Params?
      raise IOError('MOPED does not currently support sweep values option')
    elif mode == 'FixedValue':
      self.raiseADebug(f'Building pyomo capacity parameter for '
                       f'{comp.name}')
      # Params represent constant value components of the problem
      value = element._capacity._vp._parametric
      # Considering user inputs for default heron sign convention
      if value < 1:
        value *= -1
      value *= capacity_mult
      param = pyo.Param(initialize=value)
      setattr(self._m, f'{comp.name}', param)
    elif mode == 'SyntheticHistory':
      self.raiseADebug(f'Building capacity with synthetic histories for '
                       f'{comp.name}')
      # This method runs external ROM loader and defines some pyomo sets
      capacity = self.loadSyntheticHistory(element._capacity._vp._var_name, capacity_mult)
      # TODO how to better handle capacities based on Synth Histories
      self._component_meta[comp.name]['Capacity'] = capacity
    if mode != 'SyntheticHistory':
      # TODO smarter way to do this check?
      self._component_meta[comp.name]['Capacity'] = getattr(self._m, f'{comp.name}')
    if kind == 'produces':
      # TODO should we handle transfer functions here?
      for con in element._consumes:
        transfer_values = element.get_transfer().get_coefficients()
        self._component_meta[comp.name]['Consumes'] = con
        self._component_meta[comp.name]['Transfer'] = abs(transfer_values[con]) / abs(transfer_values[element._produces[0]])

  def buildCashflowMeta(self):
    """
      Builds cashflow meta used in cashflow component construction
      @ In, None
      @ Out, None
    """
    # NOTE assumes that each component can only have one cap, yearly, and repeating
    for comp in self._components:
      self.raiseADebug(f'Retrieving cashflow information for {comp.name}')
      self._cf_meta[comp.name] = {}
      self._cf_meta[comp.name]['Lifetime'] = comp._economics._lifetime
      for cf in comp._economics._cash_flows:
        # This is used later in cashflow object generation, can be unique to each cf a comp has
        params = {'tax':cf._taxable,
                  'inflation':cf._inflation,
                  'mult_target':cf._mult_target,
                  'reference':cf._reference.get_value(),
                  'X':cf._scale.get_value(),
                  }
        if params['inflation'] == 'none':
          params['inflation'] = None
        multiplier = cf._driver._multiplier
        driver_type = cf._driver.type
        # Default mult should be 1
        if multiplier == None:
          multiplier = 1
        # This corrects sign for MOPED from user inputs for demanding cashflows
        # Allows MOPED and default HERON to follow same sign conventions for inputs
        if len(comp._demands) > 0:
          multiplier *= -1
        # Using reference prices for cashflows, considering uncertain market prices
        if cf._alpha.type == 'SyntheticHistory':
          signal = cf._alpha._vp._var_name
          alpha = self.loadSyntheticHistory(signal, multiplier)
        else:
          alpha = cf._alpha._vp._parametric * multiplier
        if cf._type == 'one-time':
          # TODO consider other driver types
          if driver_type == 'FixedValue':
            self._cf_meta[comp.name]['Capex Driver'] = cf._driver._vp._parametric
          else:
            self._cf_meta[comp.name]['Capex Driver'] = None
          self._cf_meta[comp.name]['Capex'] = alpha
          self._cf_meta[comp.name]['Capex Params'] = params
          # Necessary if capex has depreciation and amortization
          self._cf_meta[comp.name]['Deprec'] = cf._depreciate
        elif cf._type == 'repeating':
          if cf._period == 'year':
            if driver_type == 'FixedValue':
              self._cf_meta[comp.name]['Yearly Driver'] = cf._driver._vp._parametric
            else:
              self._cf_meta[comp.name]['Yearly Driver'] = None
            self._cf_meta[comp.name]['Yearly'] = alpha
            self._cf_meta[comp.name]['Yearly Params'] = params
            continue
          if driver_type == 'FixedValue':
            self._cf_meta[comp.name]['Dispatch Driver'] = cf._driver._vp._parametric
          else:
            self._cf_meta[comp.name]['Dispatch Driver'] = None
          self._cf_meta[comp.name]['Dispatching'] = alpha
          self._cf_meta[comp.name]['Dispatching Params'] = params

  def createCapex(self, comp, alpha, capacity, unique_params):
    """
      Builds capex TEAL cashflow for a given component
      @ In, comp, TEAL component object
      @ In, alpha, float, reference price for capex cost
      @ In, capacity, pyomo var, size of the ocmponent that drives the cost
      @ In, unique_params, dict, settings for inflation, tax, and mult for cf
      @ Out, cf, TEAL cashflow
    """
    life = comp.getLifetime()
    cf = CashFlows.Capex()
    cf.name = 'Capex'
    cf.initParams(life)
    cfParams = {'name': 'Capex',
                'alpha': alpha,
                'driver': capacity,
                'reference': unique_params['reference'],
                'X': unique_params['X'],
                'mult_target': unique_params['mult_target'],
                }
    cf.setParams(cfParams)
    return cf

  def createRecurringYearly(self, comp, alpha, driver, unique_params):
    """
      Constructs the parameters for capital expenditures
      @ In, comp, TEAL.src.CashFlows.Component, main structure to add component cash flows
      @ In, alpha, float, yearly price to populate
      @ In, driver, pyomo.core.base.var.ScalarVar, quantity sold to populate
      @ In, unique_params, dict, settings for inflation, tax, and mult for cf
      @ Out, cf, TEAL.src.CashFlows.Component, cashflow sale for the recurring yearly
    """
    # Necessary to make life integer valued for numpy
    life = int(self._case._global_econ['ProjectTime'])
    cf = CashFlows.Recurring()
    cfParams = {'name': 'Yearly',
                'X': unique_params['X'],
                'mult_target': unique_params['mult_target'],
                }
    cf.setParams(cfParams)
    # 0 for first year (build year) -> TODO couldn't this be automatic?
    alphas = np.ones(life + 1, dtype=object) * alpha
    drivers = np.ones(life + 1, dtype=object) * driver
    alphas[0] = 0
    drivers[0] = 0
    # construct annual summary cashflows
    cf.computeYearlyCashflow(alphas, drivers)
    return cf

  def createRecurringHourly(self, comp, alpha, driver, real, unique_params):
    """
      Generates recurring hourly cashflows, mostly for dispatch and sales
      @ In, comp, TEAL component
      @ In, alpha, float/np.array, reference price of sale
      @ In, driver, numpy array of pyomo.var.values that drive cost
      @ In, real, int, current realization number
      @ In, unique_params, dict, settings for inflation, tax, and mult for cf
      @ Out, cf, TEAL cashflow
    """
    # Necessary to make integer for numpy arrays
    life = int(self._case._global_econ['ProjectTime'])
    cf = CashFlows.Recurring()
    cfParams = {'name': f'Dispatching_{real+1}',
                'X': unique_params['X'],
                'mult_target': unique_params['mult_target'],
                }
    cf.setParams(cfParams)
    cf.initParams(life, pyomoVar=True)
    # Necessary to shift year index by one since no recurring cashflows on first build year
    for year in range(life + 1):
      # Alpha can be a fixed single value price or an array of prices for each timestep
      if isinstance(alpha, float):
        cf.computeIntrayearCashflow(year, alpha, driver[year, :])
      else:
        cf.computeIntrayearCashflow(year, alpha[year, :], driver[year, :])
    return cf

  def collectResources(self):
    """
      Searches through components to collect all resources into a list
      @ In, None
      @ Out, None
    """
    for comp in self._components:
      for prod in comp._produces:
        if prod._capacity_var not in self._resources:
          self._resources.append(prod._capacity_var)
          # TODO add for consuming components
      for dem in comp._demands:
        resource = dem._capacity_var
        if resource not in self._resources:
            self._resources.append(resource)

  def buildMultiplicityVariables(self):
    """
      Generates pyomo params for applying multiplicity to dispatch vars/params
      @ In, None
      @ Out, None
    """
    if self._eval_mode == 'clustered':
      self.raiseADebug('Building multiplicity vector for clustered ROM evaluation...')
    else:
      self.raiseADebug('Building multiplicity filler for full ROM evaluation...')
    project_life = int(self._case._global_econ['ProjectTime'])
    for year in range(project_life):
      # Multiplicity used to scaled dispatches based on cluster and year
      mult = pyo.Param(self._m.c, self._m.t,
                          initialize=lambda m, c, t: self._multiplicity_meta[year+1][c],
                          domain=pyo.NonNegativeReals
                          )
      setattr(self._m, f'multiplicity_{year+1}',mult)

  def buildDispatchVariables(self, comp):
    """
      Generates dispatch vars and value arrays to build components
      @ In, comp, HERON component
      @ Out, capacity, np.array/pyomo.var, capacity variable for the component
      @ Out, template_array, np.array, array of pyo.values used for TEAL cfs
    """
    # NOTE Assumes that all components will remain functional for project life
    project_life = int(self._case._global_econ['ProjectTime'])
    # Necessary to make year index one larger than project life so that year zero
    # Can be empty for recurring cashflows
    template_array = np.zeros((self._case._num_samples, project_life + 1, self._yearly_hours), dtype=object)
    capacity = self._component_meta[comp.name]['Capacity']
    dispatch_type = self._component_meta[comp.name]['Dispatch']
    # What to have user be able to define consuming components capacity in terms of either resource
    # This allows dispatch to be in terms of producing resource and capacity be in terms of consumption resource
    # Only applies to fixed dispatch here due to pyomo construction
    if self._component_meta[comp.name]['Consumes'] is not None:
      if self._component_meta[comp.name]['Consumes'] == self._component_meta[comp.name]['Capacity Resource']:
        reverse_transfer = 1 / self._component_meta[comp.name]['Transfer']
    else:
      reverse_transfer = 1
    # Checking for type of capacity is necessary to build dispatch variable
    self._m.dummy = pyo.Var()
    self._m.placeholder = pyo.Param()
    dummy_type = type(self._m.dummy)
    placeholder_type = type(self._m.placeholder)
    self.raiseADebug(f'Preparing dispatch container for {comp.name}...')
    for real in range(self._case._num_samples):
      for year in range(project_life):
        mult = getattr(self._m,f'multiplicity_{year+1}')
        # TODO account for other variations of component settings, specifically if dispatchable
        if isinstance(capacity, (dummy_type, placeholder_type)):
          # Currently independent and dependent are interchangable
          if dispatch_type in ['independent', 'dependent']:
            var = pyo.Var(self._m.c, self._m.t,
                          initialize=lambda m, c, t: 0,
                          domain=pyo.NonNegativeReals
                          )
            setattr(self._m, f'{comp.name}_dispatch_{real+1}_{year+1}', var)
            # Shifting index such that year 0 remains 0
            # Weighting each dispatch by the number of realizations (equal weight for each realization)
            # This corrects the NPV value
            template_array[real, year + 1, :] = (1 / self._case._num_samples) * np.array(list(var.values())) * np.array(list(mult.values()))
          elif dispatch_type == 'fixed':
            param = pyo.Var(self._m.c, self._m.t,
                            initialize=lambda m, c, t: capacity.value,
                            domain=pyo.NonNegativeReals,)
            setattr(self._m, f'{comp.name}_dispatch_{real+1}_{year+1}', param)
            con = pyo.Constraint(self._m.c, self._m.t, expr=lambda m, c, t: param[(c, t)] == reverse_transfer*capacity)
            setattr(self._m, f'{comp.name}_fixed_{real+1}_{year+1}', con)
            template_array[real, year + 1, :] = (1 / self._case._num_samples) * np.array(list(param.values())) * np.array(list(mult.values()))
        else:
          if dispatch_type in ['independent', 'dependent']:
            var = pyo.Var(self._m.c, self._m.t,
                          initialize=lambda m, c, t: 0,
                          domain=pyo.NonNegativeReals,
                          bounds=lambda m, c, t: (0, capacity[f'Realization_{real+1}'][year, c, t])
                          )
            setattr(self._m, f'{comp.name}_dispatch_{real+1}_{year+1}', var)
            template_array[real, year + 1, :] = (1 / self._case._num_samples) * np.array(list(var.values())) * np.array(list(mult.values()))
          elif dispatch_type == 'fixed':
            param = pyo.Param(self._m.c, self._m.t,
                              initialize=lambda m, c, t: reverse_transfer*capacity[f'Realization_{real+1}'][year, c, t]
                              )
            setattr(self._m, f'{comp.name}_dispatch_{real+1}_{year+1}', param)
            template_array[real, year + 1, :] = (1 / self._case._num_samples) * np.array(list(param.values())) * np.array(list(mult.values()))
    return capacity, template_array

  def buildConsumptionVariables(self, comp):
    """
      Builds consumption pyomo variables that are dependent on the output of the same components dispatch
      @ In, comp, HERON component object
      @ Out, None
    """
    # NOTE Assumes that all components will remain functional for project life
    project_life = int(self._case._global_econ['ProjectTime'])
    transfer = self._component_meta[comp.name]['Transfer']
    for real in range(self._case._num_samples):
      for year in range(project_life):
        dispatch = getattr(self._m,f'{comp.name}_dispatch_{real+1}_{year+1}')
        var = pyo.Var(self._m.c, self._m.t,
                      initialize=lambda m, c, t: 0,
                      domain=pyo.NonNegativeReals)
        setattr(self._m,f'{comp.name}_consume_{real+1}_{year+1}',var)
        con = pyo.Constraint(self._m.c,self._m.t,
                             rule=lambda m, c, t: var[(c,t)] == transfer*dispatch[(c,t)])
        setattr(self._m,f'{comp.name}_consumption_limit_{real+1}_{year+1}',con)

  def buildStorageVariables(self, comp):
    """
      Builds storage dispatch(charge/discharge), level, and dependencies in pyomo
      @ In, comp, HERON component object
      @ Out, capacity, np.array/pyomo.var, capacity variable for the component
      @ Out, template_array, np.array, array of pyo.values used for TEAL cfs
    """
    self.raiseADebug(f'Preparing storage variables for {comp.name}')
    # NOTE Assumes that all components will remain functional for project life
    project_life = int(self._case._global_econ['ProjectTime'])
    # Necessary to make year index one larger than project life so that year zero
    # Can be empty for recurring cashflows
    template_array = np.zeros((self._case._num_samples, project_life + 1, self._yearly_hours), dtype=object)
    capacity = self._component_meta[comp.name]['Capacity']
    # NOTE we assume independent for all storage components
    dispatch_type = self._component_meta[comp.name]['Dispatch']
    initial_value = self._component_meta[comp.name]['Initial Value']
    trip_efficiency = self._component_meta[comp.name]['SRTE']
    # TODO how to dynamically generate the time-step value?
    dt = self._m.t[2] - self._m.t[1]
    cluster_end = self._m.t[-1]
    for real in range(self._case._num_samples):
      for year in range(project_life):
        mult = getattr(self._m,f'multiplicity_{year+1}')
        # battery needs to track level, charging, and discharging
        level = pyo.Var(self._m.c, self._m.t,
                        domain=pyo.NonNegativeReals)
        setattr(self._m,f'{comp.name}_level_{real+1}_{year+1}',level)
        level_upper = pyo.Constraint(self._m.c, self._m.t,
                                     rule=lambda m, c, t: level[(c,t)] <= capacity)
        setattr(self._m,f'{comp.name}_level_upper_{real+1}_{year+1}',level_upper)
        charge = pyo.Var(self._m.c, self._m.t,
                         domain=pyo.NonNegativeReals)
        setattr(self._m,f'{comp.name}_charge_{real+1}_{year+1}',charge)
        discharge = pyo.Var(self._m.c, self._m.t,
                         domain=pyo.NonNegativeReals)
        setattr(self._m,f'{comp.name}_discharge_{real+1}_{year+1}',discharge)
        discharge_limit = pyo.Constraint(self._m.c, self._m.t,
                                         rule=lambda m, c, t:
                                         discharge[(c,t)] <= trip_efficiency*self.getPreviousIndex(comp.name, level, real+1, year+1, c, t, initial_value))
        setattr(self._m,f'{comp.name}_discharge_limit_{real+1}_{year+1}',discharge_limit)
        # level is time dependent and requires propagation via constraints
        level_propagation = pyo.Constraint(self._m.c, self._m.t,
                                           rule=lambda m, c, t:
                                           level[(c,t)] == self.getPreviousIndex(comp.name, level, real+1, year+1, c, t, initial_value) +
                                           dt*(trip_efficiency*charge[(c,t)] - (1/trip_efficiency)*discharge[(c,t)]))
        setattr(self._m,f'{comp.name}_level_propagation_{real+1}_{year+1}',level_propagation)
        # Storage set points should enforce shorter time horizons for storage decisions
        level_point_set_lower = pyo.Constraint(self._m.c,
                                         rule=lambda m, c: level[(c,0)] == initial_value)
        setattr(self._m,f'{comp.name}_level_setpoint_lower_{real+1}_{year+1}',level_point_set_lower)
        level_point_set_upper = pyo.Constraint(self._m.c,
                                         rule=lambda m, c: level[(c,cluster_end)] == initial_value)
        setattr(self._m,f'{comp.name}_level_setpoint_upper_{real+1}_{year+1}',level_point_set_upper)
        # TODO currently only considering costs associated with discharging the storage, however charging and level should be considered
        # This will involve handling a separate template_array as a driver for a separate TEAL cashflow
        template_array[real, year+1, :] = (1 / self._case._num_samples) * np.array(list(discharge.values())) * np.array(list(mult.values()))
    return capacity, template_array

  def createCashflowComponent(self, comp, capacity, dispatch):
    """
      Builds TEAL component using pyomo dispatch and capacity variables
      @ In, capacity, pyomo.var/pyomo.param, primary driver
      @ In, life, int, number of years the component operates without replacement
      @ In, dispatch, np.array, pyomo values for dispatch variables
      @ Out, component, TEAL.Component
    """
    # Need to have TEAL component for cashflow functionality
    component = CashFlows.Component()
    params = {'name': comp.name}
    cfs = []
    cf_meta = self._cf_meta[comp.name]
    # Using read meta to evaluate possible cashflows
    for cf, value in cf_meta.items():
      if cf == 'Lifetime':
        self.raiseADebug(f'Setting component lifespan for {comp.name}')
        params['Life_time'] = value
        component.setParams(params)
      elif cf == 'Capex':
        # Capex is the most complex to handle generally due to amort
        self.raiseADebug(f'Generating Capex cashflow for {comp.name}')
        capex_params = cf_meta['Capex Params']
        capex_driver = cf_meta['Capex Driver']
        if capex_driver is None:
          capex = self.createCapex(component, value, capacity, capex_params)
        else:
          capex = self.createCapex(component, value, capex_driver, capex_params)
        cfs.append(capex)
        depreciation = cf_meta['Deprec']
        if depreciation is not None:
          capex.setAmortization('MACRS', depreciation)
          amorts = component._createDepreciation(capex)
          cfs.extend(amorts)
      # Necessary to avoid error message from expected inputs
      elif cf in ['Deprec', 'Capex Driver', 'Yearly Driver', 'Dispatch Driver', 'Capex Params', 'Yearly Params', 'Dispatching Params']:
        continue
      elif cf == 'Yearly':
        self.raiseADebug(f'Generating Yearly OM cashflow for {comp.name}')
        yearly_params = cf_meta['Yearly Params']
        yearly_driver = cf_meta['Yearly Driver']
        if yearly_driver is None:
          yearly = self.createRecurringYearly(component, value, capacity, yearly_params)
        else:
          yearly = self.createRecurringYearly(component, value, yearly_driver, yearly_params)
        cfs.append(yearly)
      elif cf == 'Dispatching':
        # Here value can be a np.array as well for ARMA grid pricing
        self.raiseADebug(f'Generating dispatch OM cashflow for {comp.name}')
        dispatching_params = cf_meta['Dispatching Params']
        dispatch_driver = cf_meta['Dispatch Driver']
        if dispatch_driver is None:
          if isinstance(value, dict):
            for real in range(self._case._num_samples):
              alpha_realization = self.reshapeAlpha(value)
              var_om = self.createRecurringHourly(component, alpha_realization[real, :, :], dispatch[real, :, :], real, dispatching_params)
              cfs.append(var_om)
          else:
            # Necessary to create a unique cash flow for each dispatch realization
            for real in range(self._case._num_samples):
              var_om = self.createRecurringHourly(component, value, dispatch[real, :, :], real, dispatching_params)
              cfs.append(var_om)
        else:
          raise IOError('MOPED does not currently handle non activity drivers for dispatch recurring cashflows')
      else:
        raise IOError(f'Unexpected cashflow type received: {cf}')
    component.addCashflows(cfs)
    return component

  def reshapeAlpha(self, alpha):
    """
      Reshapes synthetic history reference prices to match array shape of driver arrays
      @ In, alpha, dict, dictionary of numpy arrays
      @ Out, reshaped_alpha, numpy array, same data in new shape
    """
    project_life = int(self._case._global_econ['ProjectTime'])
    # plus 1 to year term to allow for 0 recurring costs during build year
    reshaped_alpha = np.zeros((self._case._num_samples,project_life+1,self._yearly_hours))
    for real in range(self._case._num_samples):
      # it necessary to have alpha be [real,year,hour] instead of [real,year,cluster,hour]
      realized_alpha = np.hstack([alpha[f'Realization_{real+1}'][:,i,:] for i in range(alpha[f'Realization_{real+1}'].shape[1])])
      reshaped_alpha[real,1:,:] = realized_alpha
    # TODO effective way of checking to see if reshape was successful?
    return reshaped_alpha

  def conserveResource(self, resource, real, year, M, c, t):
    """
      Generates pyomo constraints for resource conservation
      @ In, resource, string, name of resource we are conserving
      @ In, real, int, the current realization
      @ In, year, int, the current year
      @ In, M, pyomo.ConcreteModel
      @ In, c, int, index from pyomo set self._m.c
      @ In, t, int, index from pyomo set self._m.t
      @ Out, rule, boolean expression
    """
    # Initializing production and demand trackers
    produced = 0
    demanded = 0
    consumed = 0
    charged = 0
    discharged = 0
    # Necessary to check all components involved in the analysis
    for comp in self._components:
      comp_meta = self._component_meta[comp.name]
      # Conservation constrains the dispatch decisions
      if comp_meta['Stores'] is not None:
        charge_value = getattr(self._m, f'{comp.name}_charge_{real + 1}_{year + 1}')
        discharge_value = getattr(self._m, f'{comp.name}_discharge_{real + 1}_{year + 1}')
      else:
        dispatch_value = getattr(self._m, f'{comp.name}_dispatch_{real + 1}_{year + 1}')
      if comp_meta['Consumes'] is not None:
        consumption_value = getattr(self._m,f'{comp.name}_consume_{real+1}_{year+1}')
      for key, value in comp_meta.items():
        if key == 'Produces' and value == resource:
          produced += dispatch_value[(c,t)]
        elif key == 'Demands' and value == resource:
          demanded += dispatch_value[(c,t)]
        elif key == 'Consumes' and value == resource:
          consumed += consumption_value[(c,t)]
        elif key == 'Stores' and value == resource:
          charged += charge_value[(c,t)]
          discharged += discharge_value[(c,t)]
        # TODO consider consumption and incorrect input information
    return produced + discharged == demanded + consumed + charged

  def upper(self, comp, real, year, M, c, t):
    """
      Restricts independently dispatched compononents based on their capacity
      @ In, comp, HERON comp object
      @ In, real, int, current realization
      @ In, year, int, current year
      @ In, M, pyomo model object, MOPED pyomo ConcreteModel
      @ In, c, int, index for cluster
      @ In, t, int, index for hour within cluster
      @ Out, rule, boolean expression for upper bounding
    """
    # What to have user be able to define consuming components capacity in terms of either resource
    # This allows dispatch to be in terms of producing resource and capacity be in terms of consumption resource
    if self._component_meta[comp.name]['Consumes'] is not None:
      if self._component_meta[comp.name]['Consumes'] == self._component_meta[comp.name]['Capacity Resource']:
        reverse_transfer = 1 / self._component_meta[comp.name]['Transfer']
    else:
      reverse_transfer = 1
      # This is allows for the capacity to be an upper bound and decision variable
    upper_bound = reverse_transfer*getattr(self._m, f'{comp.name}')
    dispatch_value = getattr(self._m, f'{comp.name}_dispatch_{real+1}_{year+1}')
    return dispatch_value[(c, t)] <= upper_bound

  def buildConstraints(self):
    """
      Builds all necessary constraints for pyomo object
      @ In, None
      @ Out, None
    """
    # Convert to int to make range() viable
    project_life = int(self._case._global_econ['ProjectTime'])
    # Type variables used for checking capacity type, based on pyomo vars
    # Defined as part of the self._m pyomo model
    dummy_type = type(self._m.dummy)
    placeholder_type = type(self._m.placeholder)
    self.raiseAMessage(f'Building necessary constraints for {self._case.name}')
    for real in range(self._case._num_samples):
      for year in range(project_life):
        # Separating constraints makes sense
        # Resource conservation
        for resource in self._resources:
          con = pyo.Constraint(self._m.c, self._m.t,
                               rule=partial(self.conserveResource, resource, real, year))
          setattr(self._m, f'{resource}_con_{real+1}_{year+1}', con)
        # Bounding constraints on dispatches
        for comp in self._components:
          # storage has constraints build elsewhere see (buildStorageComponents)
          if self._component_meta[comp.name]['Stores'] is not None:
            continue
          capacity = self._component_meta[comp.name]['Capacity']
          if isinstance(capacity, (dummy_type, placeholder_type)):
            con = pyo.Constraint(self._m.c, self._m.t,
                                 rule=partial(self.upper, comp, real, year))
            setattr(self._m, f'{comp.name}_upper_{real+1}_{year+1}', con)

  def solveAndDisplay(self):
    """
      Presents results of the optimization run
      @ In, None
      @ Out, None
    """
    columns = []
    values = []
    # Results provide run times and optimizer final status
    results = self._solver.solve(self._m)
    self.raiseAMessage(f'Optimizer has finished running, here are the results\n{results}')
    for comp in self._components:
      # Not all components will have a pyomo variable
      try:
        comp_print = getattr(self._m, f'{comp.name}')
        self.raiseAMessage(f'Here is the optimized capacity for {comp.name}')
        columns.append(f'{comp.name} Capacity')
        values.append(comp_print.value)
        comp_print.pprint()
      except:
        self.raiseAMessage(f'{comp.name} does not have a standard capacity')
    NPV = pyo.value(self._m.NPV)
    self.raiseAMessage(f"The final NPV is: {NPV}")
    columns.append('Expected NPV')
    values.append(NPV)
    output_data = pd.DataFrame([values], columns=columns)
    output_data.to_csv('opt_solution.csv')

  def dispatchPlot(self, real=1, year=1, cluster=0):
    """
      Plots the dispatch behavior for a given realization, year, and cluster
      @ In, real, int, realization to plot (defaults to 1)
      @ In, year, int, year to plot (defaults to 1)
      @ In, cluster, int, cluster to plot (defaults to 1)
      @ Out, None
    """
    self.raiseAMessage(f'Generating resource dispatch plots for {self._case.name}')
    time = np.array(self._m.t)
    for res in self._resources:
      plot_colors = ['green','red','blue','orange','teal','violet','brown','black','yellow']
      plt.figure(figsize=(2,1))
      main = plt.subplot(111)
      main.set_xlabel('Time')
      main.set_ylabel(f'{res} (Dispatched)')
      main.set_title(f'{res} Dispatch (Realization: {real} Year: {year} Cluster: {cluster})')
      main.set_xlim((0,time[-1]))
      for comp in self._components:
        if self._component_meta[comp.name]['Produces'] == res:
          plot_dispatch = np.zeros(len(self._m.t))
          dispatch = getattr(self._m,f'{comp.name}_dispatch_{real}_{year}')
          for t in self._m.t:
            plot_dispatch[t] = pyo.value(dispatch[(cluster,t)])
          label = f'{comp.name} Production'
          main.plot(time,plot_dispatch,label=label,color=plot_colors[0])
          plot_colors.pop(0)
        elif self._component_meta[comp.name]['Demands'] == res:
          plot_dispatch = np.zeros(len(self._m.t))
          dispatch = getattr(self._m,f'{comp.name}_dispatch_{real}_{year}')
          for t in self._m.t:
            plot_dispatch[t] = -1*pyo.value(dispatch[(cluster,t)])
          if self._component_meta[comp.name]['Dispatch'] =='fixed':
            label = f'{comp.name} Demand'
          else:
            label = f'{comp.name} Sales'
          main.plot(time,plot_dispatch,label=label,color=plot_colors[0])
          plot_colors.pop(0)
        elif self._component_meta[comp.name]['Consumes'] == res:
          plot_dispatch = np.zeros(len(self._m.t))
          dispatch = getattr(self._m,f'{comp.name}_consume_{real}_{year}')
          for t in self._m.t:
            plot_dispatch[t] = -1*pyo.value(dispatch[(cluster,t)])
          label = f'{comp.name} Consumption'
          main.plot(time,plot_dispatch,label=label,color=plot_colors[0])
          plot_colors.pop(0)
        elif self._component_meta[comp.name]['Stores'] == res:
          plot_level = np.zeros(len(self._m.t))
          plot_charge = np.zeros(len(self._m.t))
          plot_discharge = np.zeros(len(self._m.t))
          level = getattr(self._m,f'{comp.name}_level_{real}_{year}')
          charge = getattr(self._m,f'{comp.name}_charge_{real}_{year}')
          discharge = getattr(self._m,f'{comp.name}_discharge_{real}_{year}')
          for t in self._m.t:
            plot_level[t] = pyo.value(level[(cluster,t)])
            plot_charge[t] = -1*pyo.value(charge[(cluster,t)])
            plot_discharge[t] = pyo.value(discharge[(cluster,t)])
          label = f'{comp.name} Charging'
          main.plot(time,plot_charge,label=label,color=plot_colors[0],marker='x',ls='--')
          label = f'{comp.name} Discharging'
          main.plot(time,plot_discharge,label=label,color=plot_colors[0],marker='o',ls='--')
          sec_axis = main.twinx()
          sec_axis.set_ylabel(f'{res} (Stored)')
          cap = pyo.value(getattr(self._m,f'{comp.name}'))
          sec_axis.set_ylim((0,cap+1))
          label = f'{comp.name} Level'
          sec_axis.plot(time,plot_level,label=label,color=plot_colors[0],ls='-.')
          plot_colors.pop(0)
      leg_main = main.legend(loc='best')
      leg_main.set_title('Dispatch')
      leg_sec = sec_axis.legend(loc='best')
      leg_sec.set_title('Storage')
      plt.show()

  # ===========================
  # MAIN WORKFLOW
  # ===========================
  def run(self):
    """
      Runs the workflow
      @ In, None
      @ Out, None
    """
    # Settings and metas help to build pyomo problem with cashflows
    self.buildEconSettings()
    self.buildComponentMeta()
    self.buildCashflowMeta()
    self.buildMultiplicityMeta()
    self.collectResources()
    self.buildMultiplicityVariables()
    # Each component will have dispatch and cashflow associated
    for comp in self._components:
      # Storage components have their own unique set of pyomo variables
      if self._component_meta[comp.name]['Stores'] is None:
        capacity, dispatch = self.buildDispatchVariables(comp)
      else:
        capacity, dispatch = self.buildStorageVariables(comp)
      cf_comp = self.createCashflowComponent(comp, capacity, dispatch)
      if self._component_meta[comp.name]['Consumes'] is not None:
        self.buildConsumptionVariables(comp)
      self._cf_components.append(cf_comp)
    self.raiseAMessage(f'Building pyomo cash flow expression for {self._case.name}')
    # TEAL is our cost function generator here
    metrics = RunCashFlow.run(self._econ_settings, self._cf_components, {}, pyomoVar=True)
    self._m.NPV = pyo.Objective(expr=metrics['NPV'], sense=pyo.maximize)
    # Constraints need to be built for conservation and bounds of dispatch
    self.buildConstraints()
    # NOTE this currently displays just optimizer info and capacities and cost funtion
    # TODO does this need to present information about dispatches, how to do this?
    self.raiseAMessage(f'Running Optimizer...')
    self.solveAndDisplay()
    # TODO provide way for user to turn plotting on and off, defaults to off
    if self._plot:
      self.dispatchPlot()

  # ===========================
  # UTILITIES
  # ===========================
  def setInitialParams(self, case, components, sources):
    """
      Sets all attributes read from HERON input at once
      @ In, case, Cases.Case object
      @ In, components, list of Components.Component objects
      @ In, sources, list of Placeholders objects
      @ Out, None
    """
    self.setCase(case)
    self.setComponents(components)
    self.setSources(sources)
    self.messageHandler.initialize({'verbosity': self._case._verbosity,
                                    'callerLength': 18,
                                    'tagLength': 7,
                                    'suppressErrs': False, })
    self.raiseAMessage('Sucessfully set the input parameters for MOPED run')

  def setCase(self, case):
    """
      Sets the case attribute for the MOPED object
      @ In, case, Cases.Case object
      @ Out, None
    """
    self._case = case
    self.raiseADebug(f'Setting MOPED case variable to {case}')

  def setComponents(self, components):
    """
      Sets the components attribute for the MOPED object
      @ In, components, list of Components.Component objects
      @ Out, None
    """
    self._components = components
    self.raiseADebug(f'Setting MOPED components variable to {components}')

  def setSources(self, sources):
    """
      Sets the sources attribute for the MOPED object
      @ In, sources, list of Placeholders objects
      @ Out, None
    """
    self._sources = sources
    self.raiseADebug(f'Setting MOPED sources variable to {sources}')

  def setSolver(self, solver):
    """
      Sets optimizer that pyomo runs in MOPED
      @ In, string, solver to use
      @ Out, None
    """
    self._solver = SolverFactory(solver)
    self.raiseADebug(f'Set optimizer to be {solver}')

  def getTargetParams(self, target='all'):
    """
      Returns the case, components, and sources
      @ In, target, string, param to retrieve, defaults to 'all'
      @ Out, case, Cases.Case object
      @ Out, components, list of Components.Component objects
      @ Out, sources, list of Placeholder objects
    """
    case = self._case
    components = self._components
    sources = self._sources
    # TODO Expand this method to include all attributes that are useful to retrieve
    acceptable_targets = ['all', 'case', 'components', 'sources']
    if target == 'all':
      return case, components, sources
    elif target == 'case':
      return case
    elif target == 'components':
      return components
    elif target == 'sources':
      return sources
    else:
      raise IOError(f'Your {target} is not a valid attribute for MOPED.',
                    f'Please select from {acceptable_targets}')

  def getPreviousIndex(self, comp_name, variable, real, year, c, t, initial_value):
    """
      Given an indexed variable, returns the value of the same variable, but of the previous index
      This is specifically useful for the development of battery level constraints
      @ In, comp_name, string, name of component that variable belongs to
      @ In, variable, pyomo var object
      @ In, real, int, realization of the variable
      @ In, year, int, year of the variable
      @ In, c, int, current cluster index
      @ In, t, int, current time index
      @ In, initial_value, float, initial value of var prior to analysis start
      @ Out, previous_index, indexed pyomo var object
    """
    # Need to know length of cluster and time indexes for getting previous data
    time = len(self._m.t)
    cluster = len(self._m.c)
    if t == 0:
      if c == 0:
        if year == 1:
          return initial_value
        else:
          previous_level = getattr(self._m,f'{comp_name}_level_{real}_{year}')
          # Indexing pyomo vars is more direct hence the -1
          return previous_level[(cluster-1,time-1)]
      else:
        return variable[(c-1,time-1)]
    else:
      return variable[(c, t-1)]



# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
"""
  ValuedParams are the flexible input sources used in HERON. In some way
  they represent placeholder values to be evaluated at run time from a variety of sources,
  ranging from constants to synthetic histories to AI and others.
"""
# only type references here, as needed
from .ValuedParam import ValuedParam
from .Parametric import Parametric
from .ROM import ROM

# provide easy name access to module
from .Factory import factory





# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
"""
  Defines the ValuedParam entity.
  These are objects that need to return values, but come from
  a wide variety of different sources.
"""
from .ValuedParam import ValuedParam, InputData, InputTypes

# class for potentially dynamically-evaluated quantities
class SyntheticHistory(ValuedParam):
  """
    Represents a ValuedParam that takes values directly from synthetic histories
    sampled in RAVEN.
  """
  # these types represent values that do not need to be evaluated at run time, as they are determined.
  @classmethod
  def get_input_specs(cls):
    """
      Template for parameters that can take a scalar, an ARMA history, or a function
      @ In, None
      @ Out, spec, InputData, value-based spec
    """
    spec = InputData.parameterInputFactory('ARMA', contentType=InputTypes.StringType,
        descr=r"""indicates that this value will be taken from synthetically-generated signals,
              which will be provided to the dispatch at run time by RAVEN from trained models. The value
              of this node should be the name of a synthetic history generator in the
              \xmlNode{DataGenerator} node.""")
    spec.addParam('variable', param_type=InputTypes.StringType,
        descr=r"""indicates which variable coming from the synthetic histories this value should be taken from.""")
    return spec

  def __init__(self):
    """
      Constructor.
      @ In, None
      @ Out, None
    """
    super().__init__()
    self._var_name = None # name of the variable within the synth hist
    self._source_kind = 'ARMA'

  def read(self, comp_name, spec, mode, alias_dict=None):
    """
      Used to read valued param from XML input
      @ In, comp_name, str, name of component that this valued param will be attached to; only used for print messages
      @ In, spec, InputData params, input specifications
      @ In, mode, type of simulation calculation
      @ In, alias_dict, dict, optional, aliases to use for variable naming
      @ Out, needs, list, signals needed to evaluate this ValuedParam at runtime
    """
    super().read(comp_name, spec, mode, alias_dict=None)
    # aliases get used to convert variable names, notably for the cashflow's "capacity"
    if alias_dict is None:
      alias_dict = {}
    self._source_name = spec.value
    self._var_name = spec.parameterValues['variable']
    return [self._var_name]

  def evaluate(self, inputs, target_var=None, aliases=None):
    """
      Evaluate this ValuedParam, wherever it gets its data from
      @ In, inputs, dict, stuff from RAVEN, particularly including the keys 'meta' and 'raven_vars'
      @ In, target_var, str, optional, requested outgoing variable name if not None
      @ In, aliases, dict, optional, alternate variable names for searching in variables
      @ Out, value, dict, dictionary of resulting evaluation as {vars: vals}
      @ Out, meta, dict, dictionary of meta (possibly changed during evaluation)
    """
    if aliases is None:
      aliases = {}
    # set the outgoing name for the evaluation results
    key = self._var_name if not target_var else target_var
    # allow aliasing of complex variable names
    var_name = aliases.get(self._var_name, self._var_name)
    # the year/cluster indices have already been handled by the Dispatch Manager
    # -> we just need to know the time index
    t = inputs['HERON']['time_index']
    try:
      value = inputs['HERON']['RAVEN_vars'][var_name][t]
    except KeyError as e:
      self.raiseAnError(RuntimeError, f'variable "{var_name}" was not found among the RAVEN variables!')
    except IndexError as e:
      val = inputs['HERON']['RAVEN_vars'][var_name]
      self.raiseAnError(RuntimeError, f'Attempted to access variable "{var_name}" beyond the end of its length! ' +
                        f'Requested index {t} but max index is {len(val)-1}')
    return {key: value}, inputs



# Copyright 2022, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
"""
  New HERON workflow for setting up and running DISPATCHES cases
  (HE)RON (R)uns (D)ISPATCHES (HERD)
"""
import os.path as path
import json
import sys
import copy
import operator
import pandas as pd
from itertools import compress
import pyomo.environ as pyo
from pyomo.opt import SolverFactory
import numpy as np
import _utils as hutils
try:
  import ravenframework
except ModuleNotFoundError:
  path_to_raven = hutils.get_raven_loc()
  sys.path.append(path.abspath(path.join(path_to_raven, 'plugins')))
  sys.path.append(path_to_raven)
from ravenframework.utils import xmlUtils
from ravenframework.ROMExternal import ROMLoader

# Nuclear flowsheet function imports
# NOTE: these paths will change for next DISPATCHES release
from dispatches.models.nuclear_case.flowsheets.nuclear_flowsheet import build_ne_flowsheet
from dispatches.models.nuclear_case.flowsheets.nuclear_flowsheet import fix_dof_and_initialize

# Import function for the construction of the multiperiod model
from dispatches.models.nuclear_case.flowsheets.multiperiod import build_multiperiod_design

from idaes.core.solvers import get_solver

# append path with RAVEN location
path_to_raven = hutils.get_raven_loc()
sys.path.append(path.abspath(path.join(path_to_raven, 'scripts')))
sys.path.append(path.abspath(path.join(path_to_raven, 'plugins')))
sys.path.append(path_to_raven)

from TEAL.src import CashFlows
from TEAL.src import main as RunCashFlow
from HERON.src.Moped import MOPED

DISPATCHES_MODEL_COMPONENT_META={
  "Nuclear Case": {
    "pem":{ # will require some transfer function
      "Produces": 'hydrogen',
      "Consumes": 'electricity',
      "Cashflows":{
        "Capacity":{
          "Expressions": 'pem_capacity',
        },
        "Dispatch":{
          "Expressions": ['fs.pem.electricity'],
        },
      },
    },
    "h2tank":{
      "Stores": 'hydrogen',
      "Consumes": {},
      "Cashflows":{
        "Capacity":{
          "Expressions": 'tank_capacity',
          "Multiplier":  2.016e-3, # H2 Molar Mass = 2.016e-3 kg/mol
        },
      },
    },
    "h2turbine":{ # TODO: technically also consumes air, will need to revisit
      "Produces": 'electricity',
      "Consumes": 'hydrogen',
      "Cashflows":{
        "Capacity":{
          "Expressions": 'h2_turbine_capacity',
        },
        "Dispatch":{
          "Expressions": ['fs.h2_turbine.turbine.work_mechanical',
            'fs.h2_turbine.compressor.work_mechanical'],
          "Multiplier":  [-1, -1] # extra multiplier to ensure correct sign
        },
      },
    },
    "electricity_market":{
      "Demands":  'electricity',
      "Consumes": {},
      "Cashflows":{
        "Dispatch":{
          "Expressions": ['fs.np_power_split.np_to_grid_port.electricity',
            'fs.h2_turbine.turbine.work_mechanical',
            'fs.h2_turbine.compressor.work_mechanical'],
          "Multiplier":  [1e-3, -1e-6, -1e-6] # NOTE: h2 turbine is in W, convert to kW
        },
      },
    },
    "h2_market":{
      "Demands":  'hydrogen',
      "Consumes": {},
      "Cashflows":{
        "Dispatch":{
          "Expressions": ['fs.h2_tank.outlet_to_pipeline.flow_mol'],
          "Multiplier":  [7.2576] # convert 1/s to 1/hr and 2.016e-3 kg/mol -> kg/hr
        },
      },
    },
  },
}

class HERD(MOPED):
  """
    (HE)RON (R)uns (D)ISPATCHES
    Main class used for communicating between HERON and a
    DISPATCHES case (currently, only nuclear case)
  """

  def __init__(self):
    """
      Initializing HERD class by calling parent class init.
      @ In, None
      @ Out, None
    """
    # running the init for MOPED first to initialize empty params
    super().__init__()

    # extra parameters for HERD
    self._dmdl = None # Pyomo model specific to DISPATCHES (different from _m)
    self._dispatches_model_template = None # Template of DISPATCHES Model for HERON comparison
    self._dispatches_model_comp_names = None # keys of the dispatches_model_template
    self._time_index_map = ['years', 'days', 'hours'] # index map to save time sets to dict later
    self._metrics = []     # TEAL metrics, summed expressions
    self._results = None   # results from Dispatch solve
    self._num_samples = 0  # number of samples/scenarios/realizations for easier retrievability
    self._demand_meta = {} # saving demand data to separate dict (in case it is also sampled)
    self._synth_histories = {}  # nested dict of all synthetic histories

    # Testing - using LMP signals from JSON script as used in example Jupyter notebook
    #    intended years to test out (2022-2031, use same data. 2032-2041, use same data)
    self._test_mode = False
    self._test_synth_years = []  # the actual years to test out, [2022, 2032]
    self._test_proj_life = 0     # intended length of project, regardless of synth_years
    self._test_project_year_range = [] # actual years of project [2022, 2023, 2024, ...]
    self._test_map_synth_to_proj = [] # map from synth to actual years [2022, 2022, ..., 2032, ...]

  def _set_test_time_sets(self):
    """
      Sets object attributes for time sets specifically for JSON test.
      @ In, None
      @ Out, None
    """
    self._test_synth_years = [2022]
    self._test_proj_life = 20
    # range of years through intended project life (_test_synth_years contained within this set)
    #   year[0]-1 is the construction year
    self._test_project_year_range =  np.arange(self._test_synth_years[0],
                                            self._test_synth_years[0] + self._test_proj_life)
    # array map, same length as project year range but with entries in _test_synth_years
    test_synth_years = self._test_synth_years
    self._test_map_synth_to_proj = np.array([test_synth_years[sum(y>=test_synth_years) - 1]
                                            for y in self._test_project_year_range])

  def buildEconSettings(self, verbosity=0):
    """
      Builds TEAL economic settings for running cashflows
      @ In, verbosity, int or string, verbosity settings for TEAL
      @ Out, None
    """
    # checking for a specific case - testing the DISPATCHES base Nuclear Case
    if np.any([source.name == 'dispatches-test' for source in self._sources]):
      self._test_mode = True
      self._set_test_time_sets()
      # testing for 20 year project life, override because it doesnt match LMP JSON signal
      globalEcon = getattr(self._case,"_global_econ")
      globalEcon["ProjectTime"] = self._test_proj_life # some funky pointer stuff here

    # now run parent method
    super().buildEconSettings(verbosity)
    self._num_samples = getattr(self._case,"_num_samples")

  def buildComponentMeta(self):
    """
      Collect component attributes, actions into one dictionary
      @ In, None
      @ Out, None
    """
    self._m = pyo.ConcreteModel(name=self._case.name)
    # Considering all components in analysis to build a full pyomo solve
    for comp in self._components:
      self._component_meta[comp.name] = {}
      for prod in getattr(comp, "_produces"): # NOTE Cannot handle components producing mult. things
        self.get_component_action_meta(comp, prod, "Produces")
      for sto in getattr(comp, "_stores"):
        self.get_component_action_meta(comp, sto, "Stores")
      for dem in getattr(comp, "_demands"): # NOTE Cannot handle components producing mult. things
        self.get_component_action_meta(comp, dem, "Demands")

  def _get_demand_data(self):
    """
      Builds cashflow meta specifically for market demand data
      @ In, None
      @ Out, None
    """
    for comp in self._components:
      if comp.name.split('_')[-1] == 'market':
        # NOTE: only considering one resource in demand
        demand = getattr(comp, "_demands")[0]
        demand_cap  = getattr(demand, "_capacity")
        demand_type = demand_cap.type
        demand_vp   = getattr(demand_cap, "_vp")

        if demand_type == "FixedValue":
          demand_signal  = getattr(demand_vp, "_parametric") * -1

        elif demand_type == 'SyntheticHistory':
          signal = getattr(demand_vp, "_var_name")
          demand_signal = self.loadSyntheticHistory(signal, 1)

        self._demand_meta[comp.name] = {}
        self._demand_meta[comp.name]["Demand"] = demand_signal

  def get_component_action_meta(self, comp, action, action_type=None):
    """
      Checks the capacity type, dispatch type, and resources involved for each component
      to build component_meta. Repurposed from MOPED, doesn't create Pyomo objects
      @ In, comp, HERON component
      @ In, action, HERON produces/demands/stores node
      @ In, action_type, str, name of HERON component action type
      @ Out, None
    """
    # Organizing important aspects of problem for later access
    resource = getattr(action, "_capacity_var")
    capacity = getattr(action, "_capacity")
    mode     = getattr(capacity, "type")
    value    = getattr(capacity, "_vp")

    consumes = bool(getattr(action, "_consumes")) if hasattr(action, "_consumes") else False

    # Multiplier plays important role in capacity node, especially for VRE's
    capmult = getattr(capacity, '_multiplier')
    capacity_mult = 1 if capmult is None else capmult
    # capacity_mult = -1*capacity_mult if capacity_mult<0 else capacity_mult #NOTE: need this?

    # saving resource under action type, e.g. "Produces": "electricity"
    self._component_meta[comp.name][action_type] = resource
    self._component_meta[comp.name]['Consumes'] = {}
    self._component_meta[comp.name]['Dispatch'] = getattr(action, "_dispatchable")

    # save optimization parameters
    if mode in ('OptBounds', 'FixedValue'):
      self.raiseADebug(f'|Building pyomo capacity {mode} for {comp.name}|')
      self._component_meta[comp.name][mode] = getattr(value, "_parametric") * capacity_mult

    # sample synthetic histories
    elif mode == 'SyntheticHistory':
      self.raiseADebug(f'|Building pyomo parameter with synthetic histories for {comp.name}|')
      synth_hist = self.loadSyntheticHistory( getattr(value, "_var_name"), capacity_mult ) # runs external ROM load
      self._component_meta[comp.name][mode] = synth_hist

    # cannot do sweep values yet
    elif mode == 'SweepValues': # TODO Add capability to handle sweepvalues
      raise IOError('HERD does not currently support sweep values option')

    # NOTE not all producers consume
    # TODO should we handle transfer functions here?
    if consumes:
      for con in getattr(action, "_consumes"):
        self._component_meta[comp.name]['Consumes'][con] = getattr(action, "_transfer")

  def sample_from_ROM(self, signal, multiplier):
    """
      Loads synthetic history for a specified signal, from given ROM.
      @ In, signal, string, name of signal to sample
      @ In, multiplier, int/float, value to multiply synthetic history evaluations by
      @ Out, synthetic_data, dict, contains data from evaluated ROM
    """
    # NOTE self._sources[0]._var_names are the user assigned signal names in DataGenerators
    source = getattr(self, "_sources")[0]
    source_var_names = getattr(source, "_var_names")

    # check that signal name is available within data generator
    if signal not in source_var_names:
      raise IOError('The requested signal name is not available'
                    'from the synthetic history, check DataGenerators node in input')

    # Initializing ravenROMexternal object gives PATH access to xmlUtils
    target_file = getattr(source, "_target_file")
    runner = ROMLoader( binaryFileName=target_file)

    # TODO expand to change other pickledROM settings withing this method
    inp = {'scaling': [1]}
    nodes = []
    node = xmlUtils.newNode('ROM', attrib={'name': 'SyntheticHistory', 'subType': 'pickledRom'})
    node.append(xmlUtils.newNode('clusterEvalMode', text=self._eval_mode))
    nodes.append(node)
    runner.setAdditionalParams(nodes)

    # sample realizations from ROM
    synthetic_data = {}
    for real in range(self._num_samples):
      self.raiseAMessage(f'Loading synthetic history for signal: {signal}')
      name = f'Realization_{real + 1}'
      current_realization = runner.evaluate(inp)[0]
      # applying mult to each realization is easier than iteration through dict object later
      current_realization[signal] *= multiplier
      if self._eval_mode == 'full':
        # reshape so that a filler cluster index is made
        current_realization[signal] = np.expand_dims(current_realization[signal], axis=1)
      synthetic_data[name] = current_realization[signal]

    # saving index map, often looks like ["Year", "ROM Cluster", "Hour"]
    synthetic_data['indexMap'] = current_realization['_indexMap'][0][signal]

    # get time set data from ROM realizations
    years, days, hours = self._time_index_map # defined in __init__, order matters
    for ind in synthetic_data['indexMap']:
      # for year set, we truncate based on desired Project Time (28 yrs available)
      if ind.lower() in 'years':
        projLife = int( getattr(self._case, '_global_econ')['ProjectTime'] )
        synthetic_data[years] = np.array(current_realization[ind][0:projLife], dtype=int)
      elif ind.lower() in '_rom_cluster_days':
        synthetic_data[days]  = np.array(current_realization[ind] + 1, dtype=int)
      elif ind.lower() in 'timehours':
        synthetic_data[hours] = np.array(current_realization[ind] + 1, dtype=int)

    # extracting cluster info from ROM - how many days of year per cluster?
    #    location: runner.rom._segmentROM._macroSteps[2018]._clusterInfo['map']
    #    using attrgetter to get rid of "access to protected member" warning
    cluster_steps = operator.attrgetter('_segmentROM._macroSteps')(runner.rom)
    synthetic_data['weights_days'] = {}
    for year in synthetic_data[years]:
      synthetic_data['weights_days'][year] = {}
      for cluster in synthetic_data[days]:
        cluster_map = operator.attrgetter('_clusterInfo')(cluster_steps[year])['map']
        index = int(cluster-1)
        synthetic_data['weights_days'][year][cluster] = len(cluster_map[index])

    # check that evaluation mode is either clustered or full
    if self._eval_mode not in ['clustered', 'full']:
      raise IOError('Improper ROM evaluation mode detected, try "clustered" or "full".')
    return synthetic_data

  def load_synthetic_history_from_JSON(self):
    """
      Load synthetic history data specifically from a JSON file (very specific for testing)
      @ In, None
      @ Out, fullHist, dict, synthetic history with set data
    """
    # paths to LMP signal JSON within DISPATCHES
    # TODO: move a copy of this file to HERD? or convert to static history
    disp_nuc_path = path.dirname( build_multiperiod_design.__code__.co_filename )
    lmp_path  = path.abspath( path.join(disp_nuc_path, "lmp_signal.json") )

    # loading JSON data
    with open(lmp_path, encoding='utf-8') as fp:
      synth_hist = json.load(fp)

    # data is the same for 2022-2031, and 2032-2041
    #   to save on # of variables, just duplicate LMP values
    synth_years = self._test_synth_years # actual years to gather data from e.g., [2022, 2032]

    # building array of simulation years
    proj_life_range = self._test_project_year_range # actual year range for full project [2022->2041]
    set_years_map = self._test_map_synth_to_proj # array looks like [2022, 2022, ...., 2032, 2032, ...]

    # creating set of scenarios/realizations/samples
    n_scenarios = self._num_samples
    set_scenarios = list( range(n_scenarios) )

    # we have to rebuild the LMP signal, using set_years as a map for when to duplicate data
    if len(synth_years) < self._test_proj_life:
      full_hist = {}
      # looping through scenarios, re-building LMP signal if necessary
      for r in set_scenarios:
        # output dict key is actual year, input dict key is mapped year with duplicates
        full_hist[str(r)] = {str(y): synth_hist[str(r)][str(i)]
                                for i,y in zip(set_years_map, proj_life_range)}
    else:
      # same dictionary
      full_hist = synth_hist

    n_days = len(synth_hist['0']['2020'].keys())
    n_time = len(synth_hist['0']['2020']['1'].keys()) - 1
    full_hist['years'] = synth_years
    full_hist['days']  = range(1, int(n_days + 1) )
    full_hist['hours'] = range(1, int(n_time + 1) )

    return full_hist

  def loadSyntheticHistory(self, signal, multiplier):
    """
      Loads synthetic history for a specified signal, also sets yearly hours.
      Calls the parent method and restructures dictionary to match DISPATCHES format.
      @ In, signal, string, name of signal to sample
      @ Out, new_ist, dict, contains data from evaluated ROM
    """
    if signal == 'price' and multiplier == -1:
      multiplier *= -1 # undoing negative multiplier from one step above, price != demand

    # calling parent method for loading synthetic history
    test_JSON = (signal == 'dispatches-test' and self._test_mode)

    if test_JSON:
      synth_hist = self.load_synthetic_history_from_JSON()
    else: # normal extraction
      synth_hist = self.sample_from_ROM(signal, multiplier)

    synth_years, synth_days, synth_hours = [synth_hist[ind] for ind in self._time_index_map]
    synth_scenarios = range(self._num_samples)
    proj_years_range = self._test_project_year_range
    map_synth2proj = self._test_map_synth_to_proj
    # restructure the synthetic history dictionary to match DISPATCHES
    new_hist = {}
    new_hist['signals'] = {}
    # converting to dictionary that plays nice with DISPATCHES/IDAES
    if test_JSON:
      for scenario in synth_scenarios:
        new_hist['signals'][scenario] = {year: {day: {hour:
                                        synth_hist[str(scenario)][str(year)][str(day)][str(hour)]
                                                      for hour in synth_hours}
                                              for day in synth_days}
                                      for year in proj_years_range}
    else:
      for key, data in synth_hist.items():
        # assuming the keys are in format "Realization_i"
        if "Realization" in key:
          # realizations known as scenarios in DISPATCHES, index starting at 0
          k = int( key.split('_')[-1] )
          # years indexed by integer year (2020, etc.)
          # clusters and hours indexed starting at 1
          new_hist['signals'][synth_scenarios[k-1]] = {year: {day: {hour: data[y, day-1, hour-1]
                                                                    for hour in synth_hours}
                                                          for day in synth_days}
                                                    for y, year in enumerate(synth_years)}
    # save set time data for use within DISPATCHES
    new_hist["sets"] = {}
    new_hist["sets"]["synth_scenarios"] = list(synth_scenarios) # DISPATCHES wants this as a list
    new_hist["sets"]["synth_years"]  = np.unique(synth_years) # DISPATCHES wants this as a list
    new_hist["sets"]["synth_days"]   = np.unique(synth_days)  # DISPATCHES wants this as a range
    new_hist["sets"]["synth_hours"]  = np.unique(synth_hours) # DISPATCHES wants this as a range
    new_hist["sets"]["map_synth2proj"] = map_synth2proj # used only for tests
    new_hist["sets"]["proj_years_range"] = proj_years_range # used only for tests
    # getting weights_days - how many days does each cluster represent?
    if test_JSON:
      new_hist["weights_days"] = {yr: {cl: synth_hist[str(0)][str(yr)][str(cl)]["num_days"]
                                      for cl in synth_days}
                                for yr in synth_years}
    else:
      new_hist["weights_days"] = synth_hist['weights_days']

    # saving a copy to self, referred to later when adding timesets to Pyomo model
    self._synth_histories[signal] = copy.deepcopy(new_hist)
    return new_hist

  def _check_dispatches_compatibility(self):
    """
      Checks HERON components to match compatibility with available
      DISPATCHES flowsheets.
      @ In, None
      @ Out, None
    """
    # TODO: check for financial params/inputs?
    heron_comp_list = list( self._component_meta.keys() ) # current list of HERON components
    self.raiseADebug('|Checking compatibility between HERON and available DISPATCHES cases|')

    # check that HERON input file contains all components needed to run DISPATCHES case
    # using naming convention: d___ corresponds to DISPATCHES, h___ corresponds to HERON
    dispatches_model_template = copy.deepcopy(DISPATCHES_MODEL_COMPONENT_META)
    for dName, dModel in dispatches_model_template.items():
      dispatches_comp_list    = list( dModel.keys() )
      incompatible_components = [dComp not in heron_comp_list for dComp in dispatches_comp_list]

      # 1. first check: do component names match? NOTE: shouldn't really care, just easier to check
      if sum(incompatible_components) > 0:
        # print list of components missing from HERON input
        missing_comps = list(compress(dispatches_comp_list, incompatible_components))
        message  = f'HERON components do not match DISPATCHES Model: {dName}\n'
        message +=  'Components missing from HERON XML input file: '
        message += ', '.join(missing_comps)
        raise IOError(message)

      # now let's check individual component actions
      for dComp in dispatches_comp_list:
        hCompDict = self._component_meta[dComp]  # HERON component dict, same name as DISPATCHES
        #TODO: temp fix to not check for Cashflows just yet
        if 'Cashflows' in dModel[dComp].keys():
          del dModel[dComp]['Cashflows']
        dispatches_actions_list = list(dModel[dComp].keys())
        incompatible_actions = [dAction not in hCompDict.keys()
                                      for dAction in dispatches_actions_list]

        # 2. second check: do the components have the necessary actions?
        if sum(incompatible_actions) > 0:
          missing_actions = list(compress(dispatches_actions_list, incompatible_actions))
          message = f'HERON Component {dComp} is missing the follow attributes: '
          message += ', '.join(missing_actions)
          raise IOError(message)

        # 3. third check: do the HERON component resources match the DISPATCHES ones?
        mismatched_actions = []
        for dAction, dResource in dModel[dComp].items():
          hAction = hCompDict[dAction]  # HERON component's action, might be a dict or str
          if isinstance(hAction, dict):
            hResource = list(hAction.keys())[0] if hAction else {}
            mismatched_actions.append(hResource != dResource)
          else:
            mismatched_actions.append(hAction != dResource)

        if sum(mismatched_actions) > 0:
          message = f'Attributes of HERON Component {dComp} do not match DISPATCHES case: '
          message += ', '.join( list(compress(dispatches_actions_list, mismatched_actions)) )
          raise IOError(message)
      break

    self.raiseADebug(f'|HERON Case is compatible with {dName} DISPATCHES Model|')
    self._dispatches_model_template = DISPATCHES_MODEL_COMPONENT_META[dName] # NOTE: NOT using copy
    self._dispatches_model_comp_names = list(self._dispatches_model_template.keys())

  def _add_sets_to_pyomo(self):
    """
      Create new DISPATCHES Pyomo Model with available data
      @ In, None
      @ Out, None
    """
    signals = list( self._synth_histories.keys() )

    if 'dispatches-test' in signals:
      market_synthetic_history = self._synth_histories['dispatches-test']
    elif 'price' in signals:
      market_synthetic_history = self._synth_histories['price']
    elif 'Signal' in signals:
      market_synthetic_history = self._synth_histories['Signal']
    else:
      raise IOError('Signal name not found in generated synthetic history dictionary')

    # transferring information on Sets
    sets = market_synthetic_history['sets']
    self._dmdl.set_time  = np.unique(sets['synth_hours'])
    self._dmdl.set_days  = np.unique(sets['synth_days'])
    self._dmdl.set_years = np.unique(sets['synth_years'])
    self._dmdl.set_years_map = sets['map_synth2proj'] if self._test_mode else sets['synth_years']
    self._dmdl.set_scenarios = sets['synth_scenarios']

    # transferring information on LMP Synthetic History signal
    self._dmdl.LMP = market_synthetic_history['signals']

    # transferring information on weightings
    self._dmdl.weights_days = market_synthetic_history['weights_days']

    # NOTE: equal probability for all scenarios
    self._dmdl.weights_scenarios = {s:1/self._num_samples for s in range(self._num_samples)}

  def _add_additional_constraints(self, mdl):
    """
      Method to add additional constraints not included in DISPATCHES flowsheet
      @ In, mdl, Pyomo model
      @ Out, None
    """
    @mdl.Constraint(self._dmdl.set_time, self._dmdl.set_days, self._dmdl.set_years)
    def hydrogen_demand_constraint(blk, t, d, y):
      return blk.period[t, d, y].fs.h2_tank.outlet_to_pipeline.flow_mol[0] \
                <= self._demand_meta['h2_market']["Demand"] / 2.016e-3 # convert from kg to mol

  def _build_dispatches_model(self):
    """
      Builds full DISPATCHES Pyomo model
      @ In, None
      @ Out, None
    """
    # add time sets to Pyomo model from given synthetic history and desired project lifetime
    self._add_sets_to_pyomo()

    # pointing to necessary IDAES/DISPATCHES Physics models
    mdl_flowsheet = build_ne_flowsheet # pointing to the imported DISPATCHES nuclear flowsheet
    mdl_init  = fix_dof_and_initialize # pointing to the imported DISPATCHES fix/init method
    mdl_unfix = self.unfixDof # we add a method to unfix certain DoFs based on DISPATCHES jupyter notebooks

    # NOTE: within the build process, a tmp JSON file is created in wdir...
    build_multiperiod_design(self._dmdl,
                         flowsheet=mdl_flowsheet,
                         initialization=mdl_init,
                         unfix_dof=mdl_unfix,
                         unfix_dof_options={},
                         multiple_days=True,
                         multiyear=True,
                         stochastic=True,
                         verbose=False)

    # list of initialized TEAL components; filters out HERON components that don't have cash flows
    teal_components, heron_components = self._initialize_cash_flows()

    # looping through all sampled scenarios
    for s in self._dmdl.set_scenarios:
      # Build the connecting constraints
      self.build_connecting_constraints(self._dmdl.scenario[s],
                                  set_time=self._dmdl.set_time,
                                  set_days=self._dmdl.set_days,
                                  set_years=self._dmdl.set_years)

      # Hydrogen demand constraint (Divide the RHS by the molecular mass to convert kg/s to mol/s)
      self._add_additional_constraints(self._dmdl.scenario[s])

      # Append cash flow expressions
      for hComp, tComp in zip(heron_components, teal_components): #this zip might be danger
        # skip components within HERON that are NOT defined in DISPATCHES template
        if hComp.name not in self._dispatches_model_comp_names:
          continue
        # create cashflows using TEAL (capex, yearly or hourly)
        self._create_cash_flows_for_dispatches(self._dmdl.scenario[s], hComp, tComp, s)

      # compute desired metric using TEAL and storing it
      scenario_metric = RunCashFlow.run(self._econ_settings, teal_components, {}, pyomoVar=True)
      self._metrics.append( scenario_metric )
      del scenario_metric

  def _initialize_cash_flows(self):
    """
      Initialize and populate TEAL cash flows for all components
      @ In, None
      @ Out, None
    """
    teal_components  = []
    heron_components = []
    for hComp in self._components:
      cashflows = operator.attrgetter("_economics._cash_flows")(hComp)
      if cashflows == [] or hComp.name not in self._dispatches_model_comp_names:
        continue
      # blank TEAL cash flow
      teal_comp   = CashFlows.Component()
      teal_params = {"name": hComp.name} # beginning of params dict for TEAL

      self.raiseADebug(f'Setting component lifespan for {hComp.name}')
      teal_params['Life_time'] = self._cf_meta[hComp.name]['Lifetime']
      teal_comp.setParams(teal_params)
      teal_components.append(teal_comp)

      heron_components.append(hComp)
    return teal_components, heron_components

  def _create_cash_flows_for_dispatches(self, scenario, hComp, tComp, scenario_ind):
    """
      Create and populate TEAL cash flows for all components
      @ In, scenario, Pyomo model for given scenario
      @ In, hComp, HERON component dictionary
      @ In, tComp, TEAL cashflow component object
      @ In, scenario_ind, integer for scenario index
      @ Out, None
    """
    # collection of cash flows for given HERON component
    CF_collection = []
    CF_meta = self._cf_meta[hComp.name]

    # loop through all cashflows for given component
    for cf, value in CF_meta.items():
      if cf == 'Lifetime':
        continue

      if cf == 'Capex':
        # Capex is the most complex to handle generally due to amort
        self.raiseADebug(f'Generating Capex cashflow for {hComp.name}')
        capex_params = CF_meta['Capex Params']
        capex_driver = CF_meta['Capex Driver']

        # getting capacity Pyomo object if capex_driver is not defined
        if capex_driver is None:
          capex_driver, mult = self._get_capacity_from_dispatches_model(scenario,
                                              self._dispatches_model_template[hComp.name])
          # mult defaults to 1
          value *= mult
        # generate TEAL capex cashflow
        capex = self.createCapex(tComp,
                              alpha=value,
                              capacity=capex_driver,
                              unique_params=capex_params)
        CF_collection.append(capex)
        # define depreciation
        depreciation = CF_meta['Deprec']
        if depreciation is not None:
          capex.setAmortization('MACRS', depreciation)
          amorts = getattr(tComp, "_createDepreciation")(capex)
          CF_collection.extend(amorts)
        print(f"----{hComp.name}---- Capex Driver : {capex_driver}")

      elif cf == "Yearly":
        self.raiseADebug(f'Generating Yearly OM cashflow for {hComp.name}')
        yearly_params = CF_meta['Yearly Params']
        yearly_driver = CF_meta['Yearly Driver']
        if yearly_driver is None:
          # assuming that yearly fixed OM is based on capacity
          yearly_driver, mult = self._get_capacity_from_dispatches_model(scenario,
                                              self._dispatches_model_template[hComp.name])

        yearly = self.createRecurringYearly(tComp, value, yearly_driver, yearly_params)
        CF_collection.append(yearly)
        print(f"----{hComp.name}---- Yearly Driver : {yearly_driver}")

      elif cf == "Dispatching":
        # Here value can be a np.array as well for ARMA grid pricing
        self.raiseADebug(f'Generating dispatch OM cashflow for {hComp.name}')
        dispatching_params = CF_meta['Dispatching Params']
        dispatch_driver    = CF_meta['Dispatch Driver']
        if dispatch_driver is None:
          # these should return nested lists
          dispatch_driver = self._get_dispatch_from_dispatches_model(scenario, hComp,
                                        self._dispatches_model_template[hComp.name])
        # check for alpha as a time series
        if isinstance(value, dict):
          value = self.reshapeAlpha(value)
          value = value[scenario_ind, :, :]

        hourly = self.createRecurringHourly(tComp, value,
                                        dispatch_driver, scenario_ind,
                                        dispatching_params)
        CF_collection.append(hourly)

    tComp.addCashflows(CF_collection)

  def _get_capacity_from_dispatches_model(self, mdl, dComp):
    """
      Get capacity driver string from DISPATCHES dictionary and use it to extract
      correct capacity Pyomo expression from scenario model.
      @ In, mdl, Pyomo model for given scenario
      @ In, dComp, DISPATCHES component dictionary
      @ Out, None
    """
    cashflows_dict = dComp['Cashflows']
    capacity_dict  = cashflows_dict['Capacity']
    capacity_str   = capacity_dict['Expressions']

    capacity_driver = operator.attrgetter(capacity_str)(mdl)
    mult = capacity_dict['Multiplier'] if 'Multiplier' in capacity_dict.keys() else 1

    return capacity_driver, mult

  def _get_dispatch_from_dispatches_model(self, mdl, hComp, dComp):
    """
      Get dispatch driver string from DISPATCHES dictionary and use it to extract
      correct dispatch Pyomo expression from scenario model.
      @ In, mdl, Pyomo model for given scenario
      @ In, hComp, HERON component dictionary
      @ In, dComp, DISPATCHES component dictionary
      @ Out, None
    """
    # project life time + 1, first year has to be 0 for recurring cash flows
    project_life = int( operator.attrgetter("_case._global_econ")(self)['ProjectTime'] )
    project_life += 1

    # time indeces for HERON/TEAL
    n_hours = len(self._dmdl.set_time)
    n_days  = len(self._dmdl.set_days)
    n_years = len(self._dmdl.set_years)
    n_hours_per_year = n_hours * n_days # sometimes number of days refers to clusters < 365

    set_years_map = np.hstack([0, self._dmdl.set_years_map])

    # template array for holding dispatch Pyomo expressions/objects
    dispatch_array = np.zeros((project_life, n_hours_per_year), dtype=object)
    dispatch_type = self._component_meta[hComp.name]['Dispatch']

    # TODO: there should be a more robust check of dispatch type, further upstream
    if dispatch_type == "fixed":
      self.raiseADebug("Dispatch type is fixed, this should be set through fix/unfix methods")
      return None, None

    # extract Pyomo expressions from DISPATCHES objects
    cashflows_dict = dComp['Cashflows']
    dispatch_dict  = cashflows_dict['Dispatch']
    dispatch_strs  = dispatch_dict['Expressions']

    # time indeces for DISPATCHES, as array of tuples
    indeces = np.array([tuple(i) for i in mdl.period_index], dtype="i,i,i")
    time_shape = (n_years, n_hours_per_year) # reshaping the tuples array to match HERON dispatch
    indeces = indeces.reshape(time_shape)

    # extra multipliers specific to DISPATCHES (e.g., have to multiply turbine work done by -1)
    dMults = dispatch_dict['Multiplier'] \
                if 'Multiplier' in dispatch_dict.keys() \
                else np.ones(len(dispatch_strs)) # defaults to just 1
    pcount = -1
    for p,pyear in enumerate(set_years_map):
      if pyear == 0:
        continue

      if pyear > set_years_map[p-1]:
        pcount +=1

      for time in range(n_hours_per_year):
        ind = tuple(indeces[pcount,time])
        # looping through all DISPATCHES variables pertaining to this specific dispatch
        #   e.g., turbine costs due to work done by turbine + compressor, separate variables
        dispatch_driver = 0
        for ds, dStr in enumerate(dispatch_strs):
          dispatch_driver += operator.attrgetter(dStr)(mdl.period[ind])[0] * dMults[ds]

        # getting weights for each day/cluster
        dy, yr = ind[1:]
        weight = self._dmdl.weights_days[yr][dy]  # extracting weight for year + day

        # storing individual Pyomo dispatch
        dispatch_array[p, time] = dispatch_driver * weight

    return dispatch_array

  def reshapeAlpha(self, alpha):
    """
      Reshape alpha time series (multiplier for a cash flow) to match DISPATCHES structure.
      @ In, alpha, synthetic history dictionary for time series cash flow
      @ Out, None
    """
    project_life = int( operator.attrgetter("_case._global_econ")(self)['ProjectTime'] )
    project_life += 1

    signal = alpha['signals']
    set_scenarios = alpha['sets']['synth_scenarios']
    set_years = alpha['sets']['proj_years_range'] if self._test_mode else alpha['sets']['synth_years']
    set_days  = alpha['sets']['synth_days']
    set_time  = alpha['sets']['synth_hours']

    # time indeces for HERON/TEAL
    n_hours = len(set_time)
    n_days  = len(set_days)
    n_scenarios = len(set_scenarios)
    n_hours_per_year = n_hours * n_days # sometimes number of days refers to clusters < 365

    # plus 1 to year term to allow for 0 recurring costs during build year
    reshaped_alpha = np.zeros([n_scenarios, project_life, n_hours_per_year])

    for real in set_scenarios:
      # it necessary to have alpha be [real, year, hour] instead of [real, year, cluster, hour]
      realized_alpha = [[signal[real][y][d][t] \
                                    for d in set_days
                                      for t in set_time]
                                        for y in set_years] #shape here is [year, hour]
      # first column of 2nd axis is 0 for project year 0
      reshaped_alpha[real,1:,:] = realized_alpha

    return reshaped_alpha

  def unfixDof(self, ps, **kwargs):
    """
      This function unfixes a few degrees of freedom for optimization.
      This particular method is taken from the DISPATCHES jupyter notebook
      found in "dispatches/dispatches/models/nuclear_case/flowsheets"
      titled "multiperiod_design_pricetaker"
      @In, ps, Pyomo model for period within a given scenario
      @In, kwargs, extra arguments for flowsheet parameters
      @ Out, None
    """
    # Set defaults in case options are not passed to the function
    options = kwargs.get("options", {})
    air_h2_ratio = options.get("air_h2_ratio", 10.76)

    # Unfix the electricity split in the electrical splitter
    ps.fs.np_power_split.split_fraction["np_to_grid", 0].unfix()

    # Unfix the holdup_previous and outflow variables
    ps.fs.h2_tank.tank_holdup_previous.unfix()
    ps.fs.h2_tank.outlet_to_turbine.flow_mol.unfix()
    ps.fs.h2_tank.outlet_to_pipeline.flow_mol.unfix()

    # Unfix the flowrate of air to the mixer
    ps.fs.mixer.air_feed.flow_mol.unfix()

    # Add a constraint to maintain the air to hydrogen flow ratio
    ps.fs.mixer.air_h2_ratio = pyo.Constraint(
                expr=ps.fs.mixer.air_feed.flow_mol[int(0)] ==
                      air_h2_ratio * ps.fs.mixer.hydrogen_feed.flow_mol[int(0)])

    # Set bounds on variables. A small non-zero value is set as the lower
    # bound on molar flowrates to avoid convergence issues
    ps.fs.pem.outlet.flow_mol[0].setlb(0.001)

    ps.fs.h2_tank.inlet.flow_mol[0].setlb(0.001)
    ps.fs.h2_tank.outlet_to_turbine.flow_mol[0].setlb(0.001)
    ps.fs.h2_tank.outlet_to_pipeline.flow_mol[0].setlb(0.001)

    ps.fs.translator.inlet.flow_mol[0].setlb(0.001)
    ps.fs.translator.outlet.flow_mol[0].setlb(0.001)

    ps.fs.mixer.hydrogen_feed.flow_mol[0].setlb(0.001)

  def build_connecting_constraints(self, scenario, set_time, set_days, set_years):
    """
      This function declares the first-stage variables or design decisions,
      adds constraints that ensure that the operational variables never exceed their
      design values, and adds constraints connecting variables at t - 1 and t
      @ In, scenario, Pyomo model for given scenario
      @ In, set_time, list of hours in day
      @ In, set_days, list of days/clusters per year
      @ In, set_years, list of years per scenario
      @ Out, None
    """
    # Declare first-stage variables (Design decisions)
    scenario.pem_capacity = pyo.Var(within=pyo.NonNegativeReals,
                                    doc="Maximum capacity of the PEM electrolyzer (in kW)")
    scenario.tank_capacity = pyo.Var(within=pyo.NonNegativeReals,
                                     doc="Maximum holdup of the tank (in mol)")
    scenario.h2_turbine_capacity = pyo.Var(within=pyo.NonNegativeReals,
                                           doc="Maximum power output from the turbine (in W)")

    # Ensure that the electricity to the PEM elctrolyzer does not exceed the PEM capacity
    @scenario.Constraint(set_time, set_days, set_years)
    def pem_capacity_constraint(blk, t, d, y):
      return blk.period[t, d, y].fs.pem.electricity[int(0)] <= scenario.pem_capacity

    # Ensure that the final tank holdup does not exceed the tank capacity
    @scenario.Constraint(set_time, set_days, set_years)
    def tank_capacity_constraint(blk, t, d, y):
      return blk.period[t, d, y].fs.h2_tank.tank_holdup[int(0)] <= scenario.tank_capacity

    # Ensure that the power generated by the turbine does not exceed the turbine capacity
    @scenario.Constraint(set_time, set_days, set_years)
    def turbine_capacity_constraint(blk, t, d, y):
      return (
          - blk.period[t, d, y].fs.h2_turbine.turbine.work_mechanical[int(0)]
          - blk.period[t, d, y].fs.h2_turbine.compressor.work_mechanical[int(0)] <=
          scenario.h2_turbine_capacity  )

    # Connect the initial tank holdup at time t with the final tank holdup at time t - 1
    @scenario.Constraint(set_time, set_days, set_years)
    def tank_holdup_constraints(blk, t, d, y):
      if t == 1:
        # Each day begins with an empty tank
        return (
          blk.period[t, d, y].fs.h2_tank.tank_holdup_previous[int(0)] == 0  )
      else:
        # Initial holdup at time t = final holdup at time t - 1
        return (
          blk.period[t, d, y].fs.h2_tank.tank_holdup_previous[int(0)] ==
          blk.period[t - 1, d, y].fs.h2_tank.tank_holdup[int(0)]   )

  def _add_non_anticipativity_constraints(self):
    """
      Adding non-anticipativity constraints, ensuring that all capacity variables are the same
      for all scenarios.
      @ In, None
      @ Out, None
    """
    # temporary object pointing to model
    dmdl = self._dmdl

    # Add non-anticipativity constraints
    dmdl.pem_capacity = pyo.Var(within=pyo.NonNegativeReals,
                                doc="Design PEM capacity (in kW)")
    dmdl.tank_capacity = pyo.Var(within=pyo.NonNegativeReals,
                                 doc="Design tank capacity (in mol)")
    dmdl.h2_turbine_capacity = pyo.Var(within=pyo.NonNegativeReals,
                                       doc="Design turbine capacity (in W)")

    @dmdl.Constraint(dmdl.set_scenarios)
    def non_anticipativity_pem(blk, s):
      return blk.pem_capacity == blk.scenario[s].pem_capacity

    @dmdl.Constraint(dmdl.set_scenarios)
    def non_anticipativity_tank(blk, s):
      return blk.tank_capacity == blk.scenario[s].tank_capacity

    @dmdl.Constraint(dmdl.set_scenarios)
    def non_anticipativity_turbine(blk, s):
      return blk.h2_turbine_capacity == blk.scenario[s].h2_turbine_capacity

  def _add_objective(self):
    """
      Adding objective function using TEAL metrics.
      @ In, None
      @ Out, None
    """
    # scenario probability weights
    weights = self._dmdl.weights_scenarios

    # pyomo expression for full metric wtih scenario weights applied
    Metric = np.sum( weights[n]*scenario['NPV'] for n, scenario in enumerate(self._metrics) )

    # set objective
    self._dmdl.obj = pyo.Objective(expr=Metric, sense=pyo.maximize)

  def _solve_dispatches_model(self):
    """
      Solve the DISPATCHES Pyomo model.
      @ In, None
      @ Out, None
    """
    # Define the solver object. Using the default solver: IPOPT
    solver = get_solver()

    # Solve the optimization problem
    self._results = solver.solve(self._dmdl, tee=True)

  def _export_results(self):
    """
      Print and export results from optimization solution.
      @ In, None
      @ Out, None
    """
    mdl = self._dmdl
    case_name = self._case.name

    # optimal NPV and variable values
    opt_NPV = pyo.value(mdl.obj) # given in $s
    opt_PEM = pyo.value(mdl.pem_capacity) # already in kW
    opt_H2Tank = pyo.value(mdl.tank_capacity) * 2.016e-3 # convert to kg
    opt_H2Turb = pyo.value(mdl.h2_turbine_capacity) * 1e-3 # convert to kW

    print(f'Optimal NPV is ------------------- $B {opt_NPV * 1e-9} ')
    print(f'--- Optimal PEM capacity is        {opt_PEM  * 1e-3} MW')
    print(f'--- Optimal H2 Tank capacity is    {opt_H2Tank} kg')
    print(f'--- Optimal H2 Turbine capacity is {opt_H2Turb * 1e-3} MW')

    # storing and outputting optimal values to csv
    columns = []
    values  = []
    # optimal objective / metric
    columns.append('Expected NPV')
    values.append(opt_NPV)
    # optimal capacities
    columns.extend(['PEM Size', 'H2 Tank Size', 'H2 Turbine Size'])
    values.extend([opt_PEM, opt_H2Tank, opt_H2Turb])
    # outputting to csv
    output_data = pd.DataFrame([values], columns=columns)
    output_data.to_csv(f'opt_solution__{case_name}.csv')

  def run(self):
    """
      Runs the DISPATCHES workflow
      @ In, None
      @ Out, None
    """
    # original workflow from MOPED
    self.buildEconSettings()  # overloaded method
    self.buildComponentMeta() # overloaded method
    self.buildCashflowMeta()  # MOPED method
    self.collectResources()   # MOPED method (TODO: needed?)

    # new workflow for DISPATCHES
    self._check_dispatches_compatibility()
    self._get_demand_data()

    # building the Pyomo model using DISPATCHES
    self._dmdl = pyo.ConcreteModel(name=self._case.name)
    self._build_dispatches_model()
    self._add_non_anticipativity_constraints()
    self._add_objective()
    self._solve_dispatches_model()
    self._export_results()




# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
"""
  Defines the ValuedParam entity.
  These are objects that need to return values, but come from
  a wide variety of different sources and may not be valued until run time.
"""
import sys
from HERON.src import _utils as hutils
try:
  import ravenframework
except ModuleNotFoundError:
  framework_path = hutils.get_raven_loc()
  sys.path.append(framework_path)
from ravenframework.utils import InputData, InputTypes
from ravenframework.BaseClasses import MessageUser

# class for potentially dynamically-evaluated quantities
class ValuedParam(MessageUser):
  """
    These are objects that need to return values, but come from
    a wide variety of different sources and may not be valued until run time.
  """

  @classmethod
  def get_input_specs(cls, name):
    """
      Define inputs for this VP.
      @ In, name, string, name for spec (tag)
      @ Out, spec, InputData, value-based spec
    """
    spec = InputData.parameterInputFactory(name)
    return spec

  def __init__(self):
    """
      Constructor.
      @ In, name, str, name of this valued param
      @ Out, None
    """
    super().__init__()
    self.type = self.__class__.__name__ # class type, for easy checking
    self._comp = None        # component who uses this valued param
    self._source_kind = None # if taken from a source, this is the string name of the source
    self._source_name = None # name of source object (HERON tracking name, given by user)
    self._target_obj = None  # Placeholder object for this VP, if needed
    self._value = None       # None for most VP, for Parametric may be valued

  def __repr__(self) -> str:
    """
      Return Object Representation String
      @ In, None
      @ Out, None
    """
    return f"<HERON {self._source_kind}>"

  def read(self, comp_name, spec, mode, alias_dict=None):
    """
      Used to read valued param from XML input
      Also determines what "needs" from external entities exist for this entity; these are returned
      as a list so that the input handler can connect the requisite pieces.
      @ In, comp_name, str, name of component that this valued param will be attached to; only used for print messages
      @ In, spec, InputData params, input specifications
      @ In, mode, type of simulation calculation
      @ In, alias_dict, dict, optional, aliases to use for variable naming
      @ Out, needs, list, signals needed to evaluate this ValuedParam at runtime
    """
    # aliases get used to convert variable names, notably for the cashflow's "capacity"
    if alias_dict is None:
      alias_dict = {}
    # base implementation doesn't indicate any source/signal, so we return an empty list
    return []

  def crosscheck(self, interaction):
    """
      Allows for post-reading, post-crossref checking to make sure everything is in place.
      @ In, interaction, HERON.Component.Interaction, interaction that "owns" this VP
      @ Out, None
    """
    pass # optional, use in subclasses as needed

  def get_source(self):
    """
      Accessor for the source type and name of this valued param
      @ In, None
      @ Out, kind, str, identifier for the style of valued param
      @ Out, name, str, name of the source
    """
    return self._source_kind, self._source_name

  def get_fixed_value(self):
    """
      Get the value for this parametric source.
      @ In, None
      @ Out, value, None, value
    """
    return None

  def set_object(self, obj):
    """
      Set the evaluation target of this valued param (e.g., function, ARMA, etc).
      Actual object is specific to the valued param itself.
      @ In, obj, instance, evaluation target
      @ Out, None
    """
    self._target_obj = obj

  def evaluate(self, inputs, target_var=None, aliases=None):
    """
      Evaluate this ValuedParam, wherever it gets its data from
      @ In, inputs, dict, stuff from RAVEN, particularly including the keys 'meta' and 'raven_vars'
      @ In, target_var, str, optional, requested outgoing variable name if not None
      @ In, aliases, dict, optional, alternate variable names for searching in variables
      @ Out, value, dict, dictionary of resulting evaluation as {vars: vals}
      @ Out, meta, dict, dictionary of meta (possibly changed during evaluation)
    """
    self.raiseAnError(NotImplementedError, 'Overwrite the "evaluate" method in the ValuedParam strategy!')




# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
"""
  Example class for validators.
"""
import numpy as np

from ravenframework.utils import InputData, InputTypes

from .Validator import Validator

class Example(Validator):
  """
    Example class for validating dispatch decisions.
    Arbitrarily, uses percent based ramping limits
  """
  # ---------------------------------------------
  # INITIALIZATION
  @classmethod
  def get_input_specs(cls):
    """
      Set acceptable input specifications.
      @ In, None
      @ Out, specs, InputData, specs
    """
    specs = Validator.get_input_specs()
    specs.name = 'Example'
    specs.description = r"""Uses a demonstration-only validator that constrains the change
          for any resource to a constant ``delta''."""
    specs.addSub(InputData.parameterInputFactory('delta', contentType=InputTypes.FloatType,
        descr=r"""the maximum absolute change in any resource between successive time steps."""))
    specs.addSub(InputData.parameterInputFactory('tolerance', contentType=InputTypes.FloatType,
        descr=r"""the strictness with which the constraint should be enforced. Note that some small
              numerical exception is expected."""))
    return specs

  def __init__(self):
    """
      Constructor.
      @ In, None
      @ Out, None
    """
    self.name = 'BaseValidator'
    self._allowable = 0.5
    self._tolerance = 1e-14

  def read_input(self, inputs):
    """
      Loads settings based on provided inputs
      @ In, inputs, InputData.InputSpecs, input specifications
      @ Out, None
    """
    delta = inputs.findFirst('delta')
    if delta:
      self._allowable = delta.value
    tol = inputs.findFirst('tolerance')
    if tol:
      self._tolerance = tol.value

  # ---------------------------------------------
  # API
  def validate(self, components, dispatch, times, meta):
    """
      Method to validate a dispatch activity.
      @ In, components, list, HERON components whose cashflows should be evaluated
      @ In, activity, DispatchState instance, activity by component/resources/time
      @ In, times, np.array(float), time values to evaluate; may be length 1 or longer
      @ In, meta, dict, extra information pertaining to validation
      @ Out, errs, list, information about validation failures
    """
    errs = [] # TODO best format for this?
    for comp, info in dispatch._resources.items():
      for tracker in comp.get_tracking_vars():
        for res in info:
          for t, time in enumerate(times):
            current = dispatch.get_activity(comp, tracker, res, time)
            if comp.get_interaction().is_type('Storage') and t == 0:
              init_level = comp.get_interaction().get_initial_level(meta)
            if t > 0:
              previous = dispatch.get_activity(comp, tracker, res, times[t-1])
              delta = current - previous
              sign = np.sign(delta)
              if abs(delta) - self._allowable > self._tolerance:
                errs.append({'msg': f'Exceeded ramp of {self._allowable} with {delta:1.8e}',
                             'limit': previous + (sign * self._allowable),
                             'limit_type': 'lower' if (sign < 0) else 'upper',
                             'component': comp,
                             'resource': res,
                             'time': time,
                             'time_index': t,
                            })

    return errs




# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
"""
  Base class for validators.
"""
from ravenframework.utils import InputData, InputTypes

class Validator:
  """
    Base class for strategies for validating dispatch decisions.
  """
  # ---------------------------------------------
  # INITIALIZATION
  @classmethod
  def get_input_specs(cls):
    """
      Set acceptable input specifications.
      @ In, None
      @ Out, specs, InputData, specs
    """
    specs = InputData.parameterInputFactory('Validator', ordered=False, baseNode=None)
    # only place specifications that are true for ALL validators here
    return specs

  def __init__(self):
    """
      Constructor.
      @ In, None
      @ Out, None
    """
    self.name = 'BaseValidator'

  def read_input(self, inputs):
    """
      Loads settings based on provided inputs
      @ In, inputs, InputData.InputSpecs, input specifications
      @ Out, None
    """
    pass

  def initialize(self, case, components, sources, **kwargs):
    """
      Initialize validator properties.
      @ In, case, Case, HERON case instance
      @ In, components, list, HERON components
      @ In, sources, list, HERON sources
      @ In, kwargs, dict, keyword arguments
      @ Out, None
    """
    pass

  # ---------------------------------------------
  # GETTER AND SETTERS
  # TODO

  # ---------------------------------------------
  # API
  # TODO make this a virtual method?
  def validate(self, case, components, sources, dispatch, meta):
    """
      Performs technoeconomic dispatch.
      @ In, case, Case, HERON case
      @ In, components, list, HERON components
      @ In, sources, list, HERON sources
      @ In, dispatch, HERON DispatchState, proposed activity
      @ In, meta, dict, extra information
      @ Out, results, list(dict), list of violations with info about each (see Validator base class)
    """
    # The return results should be a list of dictionaries
    # Each entry should include the following:
    # {msg: human-readable error message,
    #  limit: new limit to be imposed (as a float),
    #  limit_type: type of new limit (such as "upper" or "lower"; see below),
    #  component: HERON component object that violated conditions,
    #  resource: string name of resource for which violation was observed,
    #  time: float time at which violation was observed,
    #  time_index: integer index of time at which violation was observed,
    # }
    #
    # limit_type:
    # This will have to evolve with time. Hopefully we can keep this list updated.
    # - 'upper': signifies an upper production limit at that time step
    # - 'lower': signifies a lower production limit at that time step
    # in the future, I imagine we'll have lower_grad and upper_grad as well, who knows what else.
    raise NotImplementedError # must be implemented by inheriting classes

  # ---------------------------------------------
  # UTILITY METHODS
  # TODO




# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
"""
  Values taken from the RAVEN variable soup (in the inner)
"""
from .ValuedParam import ValuedParam, InputData, InputTypes

# class for custom dynamically-evaluated quantities
class Variable(ValuedParam):
  """
    Represents a ValuedParam that takes value from the RAVEN inner variable set
  """

  @classmethod
  def get_input_specs(cls):
    """
      Template for parameters that can take a scalar, an ARMA history, or a function
      @ In, None
      @ Out, spec, InputData, value-based spec
    """
    spec = InputData.parameterInputFactory('variable', contentType=InputTypes.StringType,
        descr=r"""the name of the variable from inner RAVEN variables.""")
    return spec

  def __init__(self):
    """
      Constructor.
      @ In, None
      @ Out, None
    """
    super().__init__()
    self._raven_var = None # name of RAVEN variable

  def read(self, comp_name, spec, mode, alias_dict=None):
    """
      Used to read valued param from XML input
      @ In, comp_name, str, name of component that this valued param will be attached to; only used for print messages
      @ In, spec, InputData params, input specifications
      @ In, mode, type of simulation calculation
      @ In, alias_dict, dict, optional, aliases to use for variable naming
      @ Out, needs, list, signals needed to evaluate this ValuedParam at runtime
    """
    super().read(comp_name, spec, mode, alias_dict=None)
    self._raven_var = spec.value
    return [self._raven_var]

  def evaluate(self, inputs, target_var=None, aliases=None):
    """
      Evaluate this ValuedParam, wherever it gets its data from
      @ In, inputs, dict, stuff from RAVEN, particularly including the keys 'meta' and 'raven_vars'
      @ In, target_var, str, optional, requested outgoing variable name if not None
      @ In, aliases, dict, optional, alternate variable names for searching in variables
      @ Out, value, dict, dictionary of resulting evaluation as {vars: vals}
      @ Out, meta, dict, dictionary of meta (possibly changed during evaluation)
    """
    if aliases is None:
      aliases = {}
    key = self._raven_var if target_var is None else target_var
    var = aliases.get(key, self._raven_var)
    try:
      val = inputs['HERON']['RAVEN_vars'][var]
    except KeyError as e:
      msg = f'ERROR: requested variable "{var}" not found among RAVEN variables!'
      msg += '  -> Available:'
      for vn in inputs['HERON']['RAVEN_vars'].keys():
        msg += f'       {vn}'
      self.raiseAnError(RuntimeError, msg)
    return {key: float(val)}, inputs




# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
"""
  Values that are swept, optimized, or fixed in the "outer" workflow,
  so end up being constants in the "inner" workflow.
"""
from .ValuedParam import ValuedParam, InputData, InputTypes

# class for custom dynamically-evaluated quantities
class Parametric(ValuedParam):
  """
    Represents a ValuedParam that takes fixed values from parametrization in the outer.
  """

  @classmethod
  def get_input_specs(cls):
    """
      Template for parameters that can take a scalar, an ARMA history, or a function
      @ In, None
      @ Out, spec, InputData, value-based spec
    """
    spec = InputData.parameterInputFactory('Parametric', contentType=InputTypes.StringType,
        descr=r"""indicates this value should be parametrized (or fixed) in the outer run,
        but functionally act as a constant in the inner workflow.""")
    spec.addParam('debug_value', param_type=InputTypes.FloatType, required=False,
        descr=r"""provides a value to use for this entry when in Case mode has "debug" enabled.""")
    return spec

  def __init__(self):
    """
      Constructor.
      @ In, None
      @ Out, None
    """
    super().__init__()
    self._parametric = None  # (low, high) for opt, sweep values for sweep, or fixed value for fixed
    self._debug_value = None # value to use if in debug mode (if desired by user)
    # NOTE that _parametric gets FIXED in the inner runs, becoming a constant
    # NOTE I think "_parametric" may only be set in the DispatchManager currently, and possibly
    #  only for Capacities. Perhaps we need a registry for valued params that keeps track
    #  of them for setting purposes.

  def read(self, comp_name, spec, mode, alias_dict=None):
    """
      Used to read valued param from XML input
      @ In, comp_name, str, name of component that this valued param will be attached to; only used for print messages
      @ In, spec, InputData params, input specifications
      @ In, mode, type of simulation calculation
      @ In, alias_dict, dict, optional, aliases to use for variable naming
      @ Out, needs, list, signals needed to evaluate this ValuedParam at runtime
    """
    super().read(comp_name, spec, mode, alias_dict=None)
    self._parametric = spec.value
    self._debug_value = spec.parameterValues.get('debug_value', None)
    return []

  def get_value(self, debug=False):
    """
      Get the value for this parametric source.
      @ In, debug, bool, optional if True then give debug value (if any)
      @ Out, value, None, value
    """
    if debug and self._debug_value is not None:
      return self._debug_value
    else:
      return self._parametric

  def set_value(self, value):
    """
      Set the value for this parametric source.
      Usually done in the Inner to fix a sampled property
      @ In, value, float, value
      @ Out, None
    """
    self._parametric = value

  def evaluate(self, inputs, target_var=None, aliases=None):
    """
      Evaluate this ValuedParam, wherever it gets its data from
      @ In, inputs, dict, stuff from RAVEN, particularly including the keys 'meta' and 'raven_vars'
      @ In, target_var, str, optional, requested outgoing variable name if not None
      @ In, aliases, dict, optional, alternate variable names for searching in variables
      @ Out, value, dict, dictionary of resulting evaluation as {vars: vals}
      @ Out, meta, dict, dictionary of meta (possibly changed during evaluation)
    """
    data = {target_var: self._parametric}
    return data, inputs

######
# dummy classes, just for changing descriptions, but they act the same as parameteric
class FixedValue(Parametric):
  @classmethod
  def get_input_specs(cls):
    """
      make specifications
      @ In, None
      @ Out, spec, InputData, value-based spec
    """
    spec = super().get_input_specs()
    spec.name = 'fixed_value'
    spec.contentType = InputTypes.FloatType
    spec.description = r"""indicates this value should be fixed in the outer run,
                       and act as a constant in the inner workflow."""
    return spec

class OptBounds(Parametric):
  @classmethod
  def get_input_specs(cls):
    """
      make specifications
      @ In, None
      @ Out, spec, InputData, value-based spec
    """
    spec = super().get_input_specs()
    spec.name = 'opt_bounds'
    spec.contentType = InputTypes.FloatListType
    spec.description = r"""indicates this value should be optimized in the outer run,
        while acting as a constant in the inner workflow."""
    return spec

class SweepValues(Parametric):
  @classmethod
  def get_input_specs(cls):
    """
      make specifications
      @ In, None
      @ Out, spec, InputData, value-based spec
    """
    spec = super().get_input_specs()
    spec.name = 'sweep_values'
    spec.contentType = InputTypes.FloatListType
    spec.description = r"""indicates this value should be parametrically swept in the outer run,
        while acting as a constant in the inner workflow."""
    return spec



import xml.etree.ElementTree as ET
import copy

def increment(item,d):
  item = item.strip().rsplit('_',1)[0]+'_{}'.format(d)
  return item

def modifyInput(root,mod_dict):
  Samplers = root.find('Samplers')
  # mc
  mc = Samplers.find('MonteCarlo')
  if mc is not None:
    # get the amount of denoising
    denoises = mod_dict['Samplers|MonteCarlo@name:mc_arma_dispatch|constant@name:denoises']
    capacities = {}
    comps = list(x.split(':')[-1].rsplit('_',1)[0] for x in mod_dict.keys() if '_capacity' in x)
    for comp in comps:
      capacities[comp] = mod_dict['Samplers|MonteCarlo@name:mc_arma_dispatch|constant@name:{}_capacity'.format(comp)]

    mc.find('samplerInit').find('limit').text = str(denoises)
    for comp, cap in capacities.items():
      for const in mc.findall('constant'):
        if const.attrib['name'] == comp+'_capacity':
          const.text = str(cap)
          break
  return root




# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
import numpy as np
from io import StringIO

class DispatchState:
  """ utility that expresses the activity (i.e. production level) of all the components in the system """
  def __init__(self):
    """
      Constructor.
      @ In, None
      @ Out, None
    """
    self._components = None # list of HERON Component objects
    self._resources = None  # Map of resources to indices for components, as {comp.name: {res, r}}
    self._times = None      # numpy array of time values, monotonically increasing

  def initialize(self, components, resources_map, times):
    """
      Set up dispatch state to hold data
      @ In, components, list, HERON components to be stored
      @ In, resources_map, dict, map of resources to indices for each component
      @ In, times, list, float times to store activity
      @ Out, None
    """
    self._components = components
    self._resources = resources_map
    self._times = times

  def __repr__(self):
    """
      Compiles string representation of object.
      @ In, None
      @ Out, repr, str, string representation
    """
    return '<HERON generic DispatchState object>'

  def get_activity(self, comp, activity, res, time, **kwargs):
    """
      Getter for activity level.
      @ In, comp, HERON Component, component whose information should be retrieved
      @ In, activity, str, tracking variable name for activity subset
      @ In, res, string, name of resource to retrieve
      @ In, time, float, time at which activity should be provided
      @ Out, activity, float, amount of resource "res" produced/consumed by "comp" at time "time";
                              note positive is producting, negative is consuming
    """
    r = self._resources[comp][res]
    t = np.searchsorted(self._times, time) # TODO protect against value not present
    return self.get_activity_indexed(comp, activity, r, t, **kwargs)

  def set_activity(self, comp, activity, res, time, value, **kwargs):
    """
      Setter for activity level.
      @ In, comp, HERON Component, component whose information should be set
      @ In, activity, str, tracking variable name for activity subset
      @ In, res, string, name of resource to retrieve
      @ In, time, float, time at which activity should be provided
      @ In, value, float, activity level; note positive is producting, negative is consuming
      @ In, kwargs, dict, additional pass-through keyword arguments
      @ Out, None
    """
    r = self._resources[comp][res]
    t = np.searchsorted(self._times, time) # TODO protect against value not present
    self.set_activity_indexed(comp, activity, r, t, value, **kwargs)

  def get_activity_indexed(self, *args, **kwargs):
    """
      Getter for activity level, using indexes instead of values for r and t
      @ In, comp, HERON Component, component whose information should be retrieved
      @ In, r, int, index of resource to retrieve (as given by meta[HERON][resource_indexer])
      @ In, t, int, index of time at which activity should be provided
      @ In, kwargs, dict, additional pass-through keyword arguments
      @ Out, activity, float, amount of resource "res" produced/consumed by "comp" at time "time";
                              note positive is producting, negative is consuming
    """
    # to be overwritten by implementing classes
    raise NotImplementedError

  def set_activity_indexed(self, *args, **kwargs):
    """
      Getter for activity level, using indexes instead of values for r and t
      @ In, args, list, additional pass-through arguments
      @ In, kwargs, dict, additional pass-through keyword arguments
      @ Out, activity, float, amount of resource "res" produced/consumed by "comp" at time "time";
                              note positive is producting, negative is consuming
    """
    # to be overwritten by implementing classes
    raise NotImplementedError

  def create_raven_vars(self, template):
    """
      Writes out RAVEN variables as expected
      @ In, template, str, formating string for variable names (using {comp}, {res})
      @ Out, data, dict, map of raven var names to numpy array data
    """
    #template = 'Dispatch__{c}__{r}' # standardized via input
    data = {}
    for comp in self._components:
      for tracker in comp.get_tracking_vars():
        for res, r in self._resources[comp].items():
          result = np.empty(len(self._times))
          for t, time in enumerate(self._times):
            result[t] = self.get_activity_indexed(comp, tracker, r, t)
          data[template.format(comp=comp.name, tracker=tracker, res=res)] = result
    return data

# NumpyState is the nominal DispatchState implementation
class NumpyState(DispatchState):
  """ implemenatation of DispatchState using Numpy. A good nominal choice if additional functionality isn't needed. """
  def __init__(self):
    """
      Constructor.
      @ In, None
      @ Out, None
    """
    DispatchState.__init__(self)
    self._data = None # dict of numpy ND array for data

  def initialize(self, components, resources_map, times):
    """
      Set up dispatch state to hold data
      @ In, components, list, HERON components to be stored
      @ In, resources_map, dict, map of resources to indices for each component
      @ In, time, list, float times to store
      @ Out, None
    """
    DispatchState.initialize(self, components, resources_map, times)
    self._data = {}
    for comp in components:
      for tag in comp.get_tracking_vars():
        self._data[f'{comp.name}_{tag}'] = np.zeros((len(self._resources[comp]), len(times)))

  def __repr__(self):
    """
      Compiles string representation of object.
      @ In, None
      @ Out, repr, str, string representation
    """
    msg = StringIO()
    msg.write('<HERON NumpyState dispatch record: \n')
    for key, act_data in self._data.items():
      name, activity = key.split('_')
      # find corresponding component
      for comp, resources in self._resources.items():
        if comp.name == name:
          break
      resources = self._resources[comp]
      msg.write(f'   component: {name} activity: {activity}\n')
      for res, r in resources.items():
        msg.write(f'      {res}: {act_data[r]}\n')
    msg.write('END NumpyState dispatch record>')
    return msg.getvalue()

  def get_activity_indexed(self, comp, activity, r, t, **kwargs):
    """
      Getter for activity level.
      Note, if any of the arguments are "None" it is assumed that means "all"
      @ In, comp, HERON Component, component whose information should be retrieved
      @ In, activity, str, tracking variable name for activity subset
      @ In, r, int, index of resource to retrieve (as given by meta[HERON][resource_indexer])
      @ In, t, int, index of time at which activity should be provided
      @ In, kwargs, dict, additional pass-through keyword arguments
      @ Out, activity, float, amount of resource "res" produced/consumed by "comp" at time "time";
                              note positive is producting, negative is consuming
    """
    return self._data[f'{comp.name}_{activity}'][r, t]

  def set_activity_indexed(self, comp, activity, r, t, value, **kwargs):
    """
      Setter for activity level.
      @ In, comp, HERON Component, component whose information should be set
      @ In, activity, str, tracking variable name for activity subset
      @ In, res, string, name of resource to retrieve
      @ In, time, float, time at which activity should be provided
      @ In, value, float, activity level; note positive is producting, negative is consuming
      @ In, kwargs, dict, additional pass-through keyword arguments
      @ Out, None
    """
    self._data[f'{comp.name}_{activity}'][r, t] = value

  # def set_activity_vector(self, comp, tracker, res, start_time, end_time, values):
  def set_activity_vector(self, comp, res, values, tracker='production', start_idx=0, end_idx=None):
    """
      Shortcut utility for setting values all-at-once in a vector.
      @ In, comp, HERON Component, component whose information should be set
      @ In, res, string, name of resource to retrieve
      @ In, values, np.array, activity level; note positive is producting, negative is consuming
      @ In, tracker, str, optional, tracking variable name for activity subset, Default: 'production'
      @ In, start_idx, int, optional, first time index at which activity is provided, Default: 0
      @ In, end_idx, int, optional, last time index at which activity is provided, Default: None
      @ Out, None
    """
    if comp.get_interaction().is_type('Storage') and tracker == 'production':
      raise RuntimeError(f'Component "{comp.name}" is Storage and the provided tracker is "production"!')
    if end_idx is None:
      end_idx = len(self._times)

    r = self._resources[comp][res]
    self._data[f'{comp.name}_{tracker}'][r, start_idx:end_idx] = values




# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
"""
  Defines the ValuedParam entity.
  These are objects that need to return values, but come from
  a wide variety of different sources.
"""
from .ValuedParam import ValuedParam, InputData, InputTypes

# class for values based on dispatch activity
class Activity(ValuedParam):
  """
    Represents a ValuedParam that takes values from dispatching activity
  """
  # these types represent values that do not need to be evaluated at run time, as they are determined.
  @classmethod
  def get_input_specs(cls):
    """
      Template for parameters that can take a scalar, an ARMA history, or a function
      @ In, None
      @ Out, spec, InputData, value-based spec
    """
    spec = InputData.parameterInputFactory('activity', contentType=InputTypes.StringType,
        descr=r"""indicates that this value will be taken from the dispatched activity of this component.
              The value of this node should be the name of a resource that this component
              either produces or consumes. Note the sign of the activity by default will be negative
              for consumed resources and positive for produced resources.""")
    spec.addParam('tracking', param_type=InputTypes.StringType, required=False,
        descr=r"""Some kinds of components have multiple tracking variables. This attribute specifies
               which tracking variable is desired. Options are only "production" for Production and Demand
               units, and [level, charge, discharge] for Storage units. Default is the first entry listed.""")
    return spec

  def __init__(self):
    """
      Constructor.
      @ In, None
      @ Out, None
    """
    super().__init__()
    self._source_kind = 'activity'
    self._var_name = None # name of the variable within the synth hist
    self._resource = None # name of the resource whose activity should be used
    self._tracking_var = None # specific tracking variable for the component

  def read(self, comp_name, spec, mode, alias_dict=None):
    """
      Used to read valued param from XML input
      @ In, comp_name, str, name of component that this valued param will be attached to; only used for print messages
      @ In, spec, InputData params, input specifications
      @ In, mode, type of simulation calculation
      @ In, alias_dict, dict, optional, aliases to use for variable naming
      @ Out, needs, list, signals needed to evaluate this ValuedParam at runtime
    """
    super().read(comp_name, spec, mode, alias_dict=None)
    subvar = spec.parameterValues.get('tracking', None)
    self._tracking_var = subvar # NOTE this gets fixed up in the crosscheck
    # aliases get used to convert variable names, notably for the cashflow's "capacity"
    if alias_dict is None:
      alias_dict = {}
    self._resource = spec.value
    # FIXME how to confirm this component actually produces/consumes/stores this resource??
    return []

  def crosscheck(self, interaction):
    """
      Allows for post-reading, post-crossref checking to make sure everything is in place.
      @ In, interaction, HERON.Component.Interaction, interaction that "owns" this VP
      @ Out, None
    """
    # use this chance to link the interaction tracking vars
    ok_trackers = interaction.get_tracking_vars()
    if self._tracking_var is None:
      self._tracking_var = ok_trackers[0]
      self.raiseAMessage(f'Tracking variable not specified; using "{self._tracking_var}" ...')
    else:
      if self._tracking_var not in ok_trackers:
        self.raiseAnError(f'Tracking variable "{self._tracking_var}" is not one of the variables ' +
                           f'tracked by this interaction! Options are: {ok_trackers}.')
    # check that the requested resource is actually used by this interaction
    available = interaction.get_resources()
    if self._resource not in available:
      str_avail = ['"{}"'.format(a) for a in available]
      self.raiseAnError(IOError, f'Requested <activity> value from resource "{self._resource}" but "{self._resource}" ' +
      f'was not found among this Component\'s input/output resources; options are:' +
      f'{", ".join(str_avail)}')

  def evaluate(self, inputs, target_var=None, aliases=None):
    """
      Evaluate this ValuedParam, wherever it gets its data from
      @ In, inputs, dict, stuff from RAVEN, particularly including the keys 'meta' and 'raven_vars'
      @ In, target_var, str, optional, requested outgoing variable name if not None
      @ In, aliases, dict, optional, alternate variable names for searching in variables
      @ Out, value, dict, dictionary of resulting evaluation as {vars: vals}
      @ Out, meta, dict, dictionary of meta (possibly changed during evaluation)
    """
    if aliases is None:
      aliases = {}
    # set the outgoing name for the evaluation results
    key = self._var_name if not target_var else target_var
    try:
      value = inputs['HERON']['activity'][self._tracking_var][self._resource]
    except KeyError as e:
      self.raiseAnError(RuntimeError, f'Resource "{self._resource}" was not found among those produced and ' +
                        'consumed by this component!')
    return {key: value}, inputs




# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
import sys
from CashFlows import CashFlowGroup
import _utils as hutils
raven_path = hutils.get_raven_loc()
sys.path.append(raven_path)

class CashFlowUser:
  """
    Base class for objects that want to access the functionality of the CashFlow objects.
    Generally this means the CashFlowUser will have an "economics" xml node used to define it,
    and will have a group of cash flows associated with it (e.g. a "component")

    In almost all cases, initialization methods should be called as part of the inheritor's method call.
  """
  @classmethod
  def get_input_specs(cls, spec):
    """
      Collects input specifications for this class.
      Note this needs to be called as part of an inheriting class's specification definition
      @ In, spec, InputData, specifications that need cash flow added to it
      @ Out, input_specs, InputData, specs
    """
    # this unit probably has some economics
    spec.addSub(CashFlowGroup.get_input_specs())
    return spec

  def __init__(self):
    """
      Constructor
      @ In, None
      @ Out, None
    """
    self._economics = None # CashFlowGroup

  def read_input(self, specs):
    """
      Sets settings from input file
      @ In, specs, InputData params, input from user
      @ Out, None
    """
    self._economics = CashFlowGroup(self)
    self._economics.read_input(specs)

  def get_crossrefs(self):
    """
      Collect the required value entities needed for this component to function.
      @ In, None
      @ Out, crossrefs, dict, mapping of dictionaries with information about the entities required.
    """
    return self._economics.get_crossrefs()

  def set_crossrefs(self, refs):
    """
      Connect cross-reference material from other entities to the ValuedParams in this component.
      @ In, refs, dict, dictionary of entity information
      @ Out, None
    """
    self._economics.set_crossrefs(refs)

  def get_incremental_cost(self, activity, raven_vars, meta, t):
    """
      get the cost given particular activities
      @ In, activity, pandas.Series, scenario variable values to evaluate cost of
      @ In, raven_vars, dict, additional variables (presumably from raven) that might be needed
      @ In, meta, dict, further dictionary of information that might be needed
      @ In, t, int, time step at which cost needs to be evaluated
      @ Out, cost, float, cost of activity
    """
    return self._economics.incremental_cost(activity, raven_vars, meta, t)

  def get_economics(self):
    """
      Accessor for economics.
      @ In, None
      @ Out, econ, CashFlowGroup, cash flows for this cash flow user
    """
    return self._economics




# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
"""
  Class for mimicking ARMA structure
"""

import numpy as np

def run(raven, raven_dict):
  """
    Makes ARMA-like data structure, same guy every time.
    @ In, raven, object, RAVEN variables object
    @ In, raven_dict, dict, additional RAVEN information
    @ Out, None
  """
  T = 24
  Y = 26
  C = 20
  data = np.ones((Y, C, T))
  raven.TOTALLOAD = data
  im = getattr(raven, '_indexMap', [{}])[0]
  im['TOTALLOAD'] = ['YEAR', '_ROM_Cluster', 'HOUR']
  raven._indexMap = np.atleast_1d([im])
  raven.YEAR = np.arange(Y) + 2025
  raven._ROM_Cluster = np.arange(C)
  raven.HOUR = np.arange(T)



